from os import getenv
from shimoku.exceptions import ActivityError, CacheError
import unittest
from utils import initiate_shimoku

s = initiate_shimoku()
business_id: str = getenv("BUSINESS_ID")
mock: bool = getenv("MOCK") == "TRUE"
async_execution: bool = getenv("ASYNC_EXECUTION") == "TRUE"

s.set_workspace(uuid=business_id)
s.set_menu_path("test_activity")

activity_name = "test_activity"

_activity = s.activities.get_activity(name=activity_name)
if not _activity:
    _activity = s.activities.create_activity(name=activity_name)

activity_id = _activity["id"]
run_id = s.activities.create_run(uuid=activity_id)["id"]


def delete_new_activity_if_it_exists():
    """
    Deletes the activity created in the test_create_delete_get_activities test.
    """
    try:
        s.activities.delete_activity(name="new_" + activity_name)
    except ActivityError:
        pass


class TestActivity(unittest.TestCase):
    def test_get_run_settings(self):
        """
        Makes a call to the activity metadata API to get run settings.
        """
        s.activities.get_run_settings(uuid=activity_id, run_id=run_id)
        s.activities.get_run_settings(uuid=activity_id, run_id=run_id)
        s.activities.get_run_settings(name=activity_name, run_id=run_id)
        s.activities.get_run_settings(name=activity_name, run_id=run_id)

    def test_create_delete_get_activities(self):
        """
        Makes various calls to the activity metadata API to create, delete and get activities in different ways.
        """

        new_activity_name = "0_new_" + activity_name
        n_activities_before = len(s.activities.get_activities())

        activity = s.activities.create_activity(name=new_activity_name)

        new_activity_id = activity["id"]

        assert activity["name"] == new_activity_name
        assert activity["runs"] == []

        n_activities_mid = len(s.activities.get_activities())
        assert n_activities_mid == n_activities_before + 1

        s.set_menu_path(name="test_activity")

        n_activities_mid = len(s.activities.get_activities())
        assert n_activities_mid == n_activities_before + 1

        s.activities.delete_activity(uuid=new_activity_id)

        n_activities_after = len(s.activities.get_activities())

        assert n_activities_before == n_activities_after

        self.activity_doesnt_exist(new_activity_name)

    def test_update_activity(self):
        """
        Makes various calls to the activity metadata API to update activities.
        """
        new_activity_name = "1_new_" + activity_name

        n_activities_before = len(s.activities.get_activities())

        activity = s.activities.create_activity(name=new_activity_name)

        new_activity_id = activity["id"]

        assert activity["name"] == new_activity_name

        s.activities.update_activity(
            uuid=new_activity_id, new_name="updated " + new_activity_name
        )

        assert not s.activities.get_activity(name=new_activity_name)
        assert s.activities.get_activity(name="updated " + new_activity_name)
        assert s.activities.create_activity(name=new_activity_name)
        self.cant_create_activity(new_activity_name)

        s.activities.delete_activity(uuid=new_activity_id)
        s.activities.delete_activity(name=new_activity_name)

        n_activities_after = len(s.activities.get_activities())

        assert n_activities_before == n_activities_after

    def test_create_get_runs(self):
        """
        Makes various calls to the activity metadata API to create, and get runs.
        """
        new_activity_name = "2_new_" + activity_name
        activity = s.activities.create_activity(name=new_activity_name)
        new_activity_id = activity["id"]

        assert activity["name"] == new_activity_name
        assert activity["runs"] == []

        run1 = s.activities.create_run(name=new_activity_name)
        run2 = s.activities.create_run(uuid=new_activity_id)
        run3 = s.activities.create_run(name=new_activity_name)

        activity = s.activities.get_activity(uuid=new_activity_id, how_many_runs=1)

        assert activity["name"] == new_activity_name
        assert len(activity["runs"]) == 1
        assert activity["runs"][0]["id"] in [run1["id"], run2["id"], run3["id"]]

        activity = s.activities.get_activity(name=new_activity_name, how_many_runs=3)
        activities_run_ids = [run["id"] for run in activity["runs"]]

        assert activity["name"] == new_activity_name
        assert len(activity["runs"]) == 3
        assert run1["id"] in activities_run_ids
        assert run2["id"] in activities_run_ids
        assert run3["id"] in activities_run_ids

        # delete the activity and test that it doesn't exist
        s.activities.delete_activity(name=new_activity_name)

        self.activity_doesnt_exist(new_activity_name)

    def test_create_get_run_logs(self):
        """
        Makes various calls to the activity metadata API to create, and get logs.
        """
        new_activity_name = "3_new_" + activity_name
        activity = s.activities.create_activity(name=new_activity_name)
        new_activity_id = activity["id"]

        assert activity["name"] == new_activity_name
        assert activity["runs"] == []

        run = s.activities.create_run(name=new_activity_name)

        log1 = s.activities.create_run_log(
            name=new_activity_name,
            run_id=run["id"],
            message="test message 1",
            severity="INFO",
        )
        log2 = s.activities.create_run_log(
            uuid=new_activity_id,
            run_id=run["id"],
            message="test message 2",
            severity="DEBUG",
            tags={"tag1": "tag1", "tag2": "tag2"},
        )
        log3 = s.activities.create_run_log(
            name=new_activity_name,
            run_id=run["id"],
            message="test message 3",
            severity="ERROR",
            tags={"tag1": "tag1"},
        )

        activity = s.activities.get_activity(name=new_activity_name, how_many_runs=1)
        run = activity["runs"][0]
        logs = s.activities.get_run_logs(uuid=new_activity_id, run_id=run["id"])

        # test that the logs are in the correct order, they are ordered by creation time
        assert len(logs) == 3
        assert log1 == logs[0]
        assert log2 == logs[1]
        assert log3 == logs[2]

        # delete the activity and test that it doesn't exist
        s.activities.delete_activity(name=new_activity_name)

        self.activity_doesnt_exist(new_activity_name)

    def test_default_activity_settings(self):
        """
        Tests that the default settings are being set and used correctly.
        """
        new_activity_name = "4_new_" + activity_name
        # test that the default settings are being set
        settings = {"name": "test1", "value": "test1"}
        activity = s.activities.create_activity(
            name=new_activity_name, settings=settings
        )

        assert activity["settings"] == settings

        # test that the default settings persist in the API
        s.set_menu_path(name="test_activity")

        activity = s.activities.get_activity(uuid=activity["id"])
        assert activity["settings"] == settings

        # delete the activity and test that it doesn't exist
        s.activities.delete_activity(name=new_activity_name)

        self.activity_doesnt_exist(new_activity_name)

    def test_get_run_logs(self):
        """
        Makes various calls to the activity metadata API to create, and get logs.
        """
        new_activity_name = "5_new_" + activity_name
        activity = s.activities.create_activity(name=new_activity_name)
        new_activity_id = activity["id"]

        run = s.activities.create_run(name=new_activity_name)

        log1 = s.activities.create_run_log(
            name=new_activity_name,
            run_id=run["id"],
            message="test message 1",
            severity="INFO",
        )
        log2 = s.activities.create_run_log(
            uuid=new_activity_id,
            run_id=run["id"],
            message="test message 2",
            severity="DEBUG",
            tags={"tag1": "tag1", "tag2": "tag2"},
        )
        log3 = s.activities.create_run_log(
            name=new_activity_name,
            run_id=run["id"],
            message="test message 3",
            severity="ERROR",
            tags={"tag1": "tag1"},
        )

        activity = s.activities.get_activity(name=new_activity_name, how_many_runs=1)
        run = activity["runs"][0]

        # test that the logs can be retrieved by run id
        logs = s.activities.get_run_logs(uuid=new_activity_id, run_id=run["id"])
        assert len(logs) == 3
        assert log1 == logs[0]
        assert log2 == logs[1]
        assert log3 == logs[2]

        # delete the activity and test that it doesn't exist
        s.activities.delete_activity(name=new_activity_name)

        self.activity_doesnt_exist(new_activity_name)

    def test_execute_activity(self):
        """
        Executes an activity and tests that the activity and run are created, also tests that the run has a log
        and that the settings are being passed correctly. The activity must have a webhook linked for it to work.
        """
        s.activities.create_webhook(name=activity_name, url="https://www.google.com")

        run = s.activities.execute_activity(name=activity_name)
        activity = s.activities.get_activity(name=activity_name, how_many_runs=1)

        # Runs are ordered by execution time, so the first run is the most recently executed
        assert activity["runs"][0]["id"] == run["id"]
        assert len(activity["runs"][0]["logs"]) == 1 or mock

        settings = [
            {"name": "test1", "value": "test1"},
            {"name": "test2", "value": "test2"},
        ]

        run1 = s.activities.execute_activity(
            name=activity_name, run_settings=settings[0]
        )
        run2 = s.activities.execute_activity(
            name=activity_name, run_settings=settings[1]
        )
        run3 = s.activities.create_run(name=activity_name, settings=run2["id"])
        s.activities.execute_run(name=activity_name, run_id=run3["id"])

        activity = s.activities.get_activity(name=activity_name, how_many_runs=4)

        assert activity["runs"][0]["id"] == run["id"]
        assert activity["runs"][0]["settings"] == {}
        assert len(activity["runs"][0]["logs"]) == 1 or mock
        assert activity["runs"][1]["id"] == run1["id"]
        assert activity["runs"][1]["settings"] == settings[0]
        assert len(activity["runs"][1]["logs"]) == 1 or mock
        assert activity["runs"][2]["id"] == run2["id"]
        assert activity["runs"][2]["settings"] == settings[1]
        assert len(activity["runs"][2]["logs"]) == 1 or mock
        assert activity["runs"][3]["id"] == run3["id"]
        assert activity["runs"][3]["settings"] == settings[1]
        assert len(activity["runs"][3]["logs"]) == 1 or mock

    def test_button_and_form_execute_activity(self):
        """
        Creates a button in the dashboard with the capability to execute an activity.
        """
        s.set_menu_path("test_activity")

        s.plt.activity_button(
            activity_name=activity_name,
            order=0,
            label="test_button",
            cols_size=12,
            align="start",
        )
        s.plt.activity_button(
            activity_name=activity_name,
            order=1,
            label="test_button",
            cols_size=12,
            align="center",
        )
        s.plt.activity_button(
            activity_name=activity_name,
            order=2,
            label="test_button",
            rows_size=2,
            cols_size=12,
            align="right",
        )

        for i in range(6):
            s.plt.activity_button(
                activity_name=activity_name,
                order=3 + i,
                label="test_button",
                rows_size=2,
                cols_size=12 - i * 2,
                align="stretch",
                padding=f"0,{i},0,{i}",
            )

        s.plt.activity_button(
            activity_name=activity_name,
            order=9,
            label="test_button",
            cols_size=12,
            align="center",
        )

        form_sent = (
            "<head>"
            "<style>"  # Styles title
            ".component-title{height:auto; width:100%; "
            "border-radius:16px; padding:16px;"
            "display:flex; align-items:center;"
            "background-color:var(--chart-C1); color:var(--color-white);}"
            "<style>.base-white{color:var(--color-white);}</style>"
            "</head>"  # Styles subtitle
            "<div class='component-title'>"
            "<div class='big-icon-banner'></div>"
            "<div class='text-block'>"
            "<h1>The activity has been called</h1>"
            "<p class='base-white'>"
            "Activity Test</p>"
            "</div>"
            "</div>"
        )

        s.plt.set_modal(modal_name="Activity called")
        s.plt.html(html=form_sent, order=0)
        s.plt.pop_out_of_modal()

        form_groups = {
            "Personal information": [
                {
                    "mapping": "name",
                    "fieldName": "name",
                    "inputType": "text",
                },
            ]
        }

        s.plt.generate_input_form_groups(
            order=10,
            form_groups=form_groups,
            activity_name=activity_name,
            modal="Activity called",
        )
        if async_execution:
            s.activate_async_execution()

    def activity_doesnt_exist(self, _activity_name):
        with self.assertRaises(ActivityError):
            s.activities.delete_activity(name=_activity_name)

    def cant_create_activity(self, _activity_name):
        with self.assertRaises(CacheError):
            s.activities.create_activity(name=_activity_name)



from typing import TYPE_CHECKING
from shimoku.code_gen.utils_code_gen import code_gen_from_list

if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import (
        AppCodeGen,
    )
    from shimoku.api.resources.report import Report


async def code_gen_from_form(
    self: "AppCodeGen", report: "Report", properties: dict
) -> list[str]:
    """Generate code for a form report.
    :param report: report to generate code from
    :param properties: properties of the report
    :return: list of code lines
    """
    rds = (await report.get_report_data_sets())[0]
    rds_properties = rds["properties"]
    form_groups = {
        group["title"]: group["fields"] for group in rds_properties["fields"]
    }
    input_form_params = []
    if rds_properties["fields"][0].get("nextFormGroup"):
        input_form_params.append("    dynamic_sequential_show=True,")
    if rds_properties.get("variant") == "autoSend":
        input_form_params.append("    auto_send=True,")

    events_on_submit = []
    if "events" not in properties:
        properties["events"] = {"onSubmit": []}
    for event in properties["events"]["onSubmit"]:
        if event["action"] == "openModal":
            input_form_params.append(
                f'    modal="{(await self._app.get_report(event["params"]["modalId"]))["properties"]["hash"]}",'
            )
        elif event["action"] == "openActivity":
            input_form_params.append(
                f'    activity_name="{(await self._app.get_activity(event["params"]["activityId"]))["name"]}",'
            )
        else:
            events_on_submit.append(event)
    if events_on_submit:
        code_gen_on_submit_events = code_gen_from_list(events_on_submit, 4)
        input_form_params.append(
            f"    on_submit_events={code_gen_on_submit_events[0][4:]}"
        )
        input_form_params.extend(code_gen_on_submit_events[1:])

    report_params = [
        f'    order={report["order"]},',
    ]
    if report["sizeColumns"] != 12:
        report_params.append(f'    cols_size={report["sizeColumns"]},')
    if report["sizeRows"] != 1:
        report_params.append(f'    rows_size={report["sizeRows"]},')
    if report["sizePadding"] != "0,0,0,0":
        report_params.append(f'    padding="{report["sizePadding"]}",')
    return [
        "shimoku_client.plt.generate_input_form_groups(",
        f"    form_groups={form_groups},",
        *report_params,
        *input_form_params,
        ")",
    ]



import logging
import asyncio
import json

from typing import Optional, Dict, Any, List, TypeVar, Type, Set, Tuple, Union, NewType

from abc import ABC

from shimoku.utils import IN_BROWSER
from shimoku.api.client import ApiClient
from shimoku.exceptions import ResourceIdMissing, CacheError
from shimoku.execution_logger import log_error, ClassWithLogging
from dataclasses import is_dataclass, asdict

if not IN_BROWSER:
    from tqdm import tqdm

logger = logging.getLogger(__name__)

IsResource = TypeVar("IsResource", bound="Resource")
Alias = NewType("Alias", Union[str, Tuple[str, ...]])
AliasField = NewType("AliasField", Union[str, Tuple[str, ...]])


class ResourceCache(ClassWithLogging):
    _module_logger = logger

    def __init__(self, resource_class: Type[IsResource], parent: IsResource):
        self._parent = parent
        self._resource_class = resource_class
        self._resource_plural = resource_class.plural
        self._cache: Dict[str:IsResource] = {}
        self._aliases: Dict[str:str] = {}
        self._listed = False
        self._listing_lock = None

    async def list(self, limit: Optional[int] = None) -> List[IsResource]:
        """List all child resources
        :param limit: Limit the number of resources to be listed
        :return: List of resources
        """
        if not self._listing_lock:
            self._listing_lock = asyncio.Lock()

        async with self._listing_lock:
            if self._parent.api_client.cache_enabled:
                if self._listed:
                    logger.debug("Resource cache already listed")
                    return list(self._cache.values())
            else:
                await asyncio.gather(
                    *[
                        resource.update()
                        for resource in self._cache.values()
                        if "dirty" in dir(resource) and resource.dirty
                    ]
                )
                self._cache = {}
                self._aliases = {}

            self._listed = True
            endpoint: str = (
                f'{self._parent["base_url"]}{self._parent.resource_type}/'
                f'{self._parent["id"]}/{self._resource_plural}'
            )

            resources_raw: Dict = await self._parent.api_client.query_element(
                endpoint=endpoint,
                method="GET",
                limit=limit,
                elastic_supported=self._resource_class.elastic_supported,
            )
            try:
                resources = (
                    resources_raw.get("items")
                    if isinstance(resources_raw, dict)
                    else resources_raw
                )
                if not isinstance(resources, list):
                    raise AttributeError()
            except AttributeError:
                logger.warning("Resource's json should have an 'items' field")
                return []

            # If the resources have no 'id' field, they have to have only one field and it has to be a dict
            # Then the resources are mapped to that field
            if resources and resources[0].get("id") is None:
                resources_keys = [
                    k for k in list(resources[0].keys()) if not k.startswith("_")
                ]
                assert len(resources_keys) == 1
                assert isinstance(resources[0][resources_keys[0]], dict)
                resources = [resource[resources_keys[0]] for resource in resources]

            # Sort resources by id to ensure consistent ordering for alias collision resolution
            resources = sorted(resources, key=lambda x: x.get("id"))

            for resource_dict in resources:
                resource = self._resource_class(
                    db_resource=resource_dict, parent=self._parent
                )
                alias_entry = resource["alias"]
                if alias_entry:
                    if alias_entry in self._aliases:
                        continue
                    self._aliases[alias_entry] = resource_dict.get("id")
                self._cache[resource_dict.get("id")] = resource
                resource.empty_changed_params()

        self._listing_lock = None
        return list(self._cache.values())

    async def add(
        self,
        resource: Optional[IsResource] = None,
        alias: Optional[Alias] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> IsResource:
        if alias:
            await self.raise_if_alias_exists(alias)

        if not resource:
            if not alias:
                log_error(
                    logger,
                    "Alias is required when resource is not provided",
                    CacheError,
                )

            resource: IsResource = self._resource_class(
                parent=self._parent, alias=alias
            )

        if not resource["id"]:
            resource.set_params(**params) if params else None
            await resource

        resource_id = resource["id"]
        if resource.resource_type != self._resource_class.resource_type:
            log_error(
                logger,
                f"Resource {resource_id} has the wrong resource type",
                CacheError,
            )

        if resource_id in self._cache:
            if self._parent.api_client.cache_enabled:
                log_error(
                    logger,
                    f"Resource {resource_id} already exists in cache",
                    CacheError,
                )
            del self._cache[resource_id]
            for alias_key, alias_value in list(self._aliases.items()):
                if alias_value == resource_id:
                    del self._aliases[alias_key]
                    break

        alias = resource["alias"]

        if alias is not None:
            self._aliases[alias] = resource_id

        self._cache[resource_id] = resource

        return resource

    async def get(
        self, uuid: Optional[str] = None, alias: Optional[Alias] = None
    ) -> Optional[IsResource]:
        if uuid and alias:
            log_error(logger, "Only one of uuid or alias can be provided", CacheError)

        if self._parent.api_client.cache_enabled:
            if alias in self._aliases:
                uuid = self._aliases[alias]

            if uuid in self._cache:
                logger.debug(f"CACHE HIT: Resource {uuid}")
                return self._cache[uuid]

        if not uuid:
            if alias is not None:
                logger.debug(f"CACHE MISS: Alias {alias}")

                await self.list()

                if alias in self._aliases:
                    uuid = self._aliases[alias]
                    return self._cache[uuid]
                else:
                    logger.debug(f"Alias {alias} not found in cache")
                    return None

            log_error(logger, "Resource id or alias have not been provided", CacheError)

        logger.debug(f"CACHE MISS: Resource {uuid}")

        db_resource = await self._resource_class(parent=self._parent, uuid=uuid).get()
        resource = self._resource_class(parent=self._parent, db_resource=db_resource)

        return await self.add(resource=resource, alias=alias)

    async def update(
        self, uuid: Optional[str] = None, alias: Optional[Alias] = None, **params
    ) -> bool:
        resource: IsResource = await self.get(uuid=uuid, alias=alias)
        if not resource:
            return False

        if params.get("new_alias"):
            new_alias = resource.get_alias_from_other_params(params)
            await self.raise_if_alias_exists(new_alias)
            del self._aliases[resource["alias"]]
            self._aliases[new_alias] = resource["id"]
            del params["new_alias"]

        resource.set_params(**params)
        await resource.update()

        return True

    async def delete(
        self, uuid: Optional[str] = None, alias: Optional[Alias] = None
    ) -> bool:
        resource: IsResource = await self.get(uuid=uuid, alias=alias)
        if not resource:
            log_error(
                logger,
                f"Resource {alias} not found in cache, unable to delete it",
                CacheError,
            )
            return False
        uuid = resource["id"]
        await resource.delete()
        del self._cache[uuid]
        self._aliases = {k: v for k, v in self._aliases.items() if v != uuid}
        return True

    async def raise_if_alias_exists(self, alias: Alias):
        await self.list()
        if alias in self._aliases:
            if self._parent.api_client.cache_enabled:
                log_error(logger, f"Alias {alias} already exists in cache", CacheError)
            del self._aliases[alias]

    # @logging_before_and_after(logging_level=logger.debug)
    def cascade_to_dict(self) -> List[Dict[str, Any]]:
        return [resource.cascade_to_dict() for resource in self._cache.values()]

    def clear(self):
        self._cache = {}
        self._aliases = {}
        self._listed = False

    def __iter__(self):
        return iter(self._cache.items())

    def __getitem__(self, key):
        return self.get(uuid=key)

    def __len__(self):
        return len(self._cache)

    def __contains__(self, item):
        return item in self._cache

    def __delitem__(self, key):
        return self.delete(uuid=key)


class BaseResource(ClassWithLogging):
    """Base class for all the resources in the API"""

    _module_logger = logger

    def __init__(
        self,
        parent: Optional[IsResource] = None,
        uuid: Optional[str] = None,
        api_client: Optional[ApiClient] = None,
        params: Optional[Dict[str, Any]] = None,
        alias_field: Optional[AliasField] = None,
        resource_type: Optional[str] = None,
        params_to_serialize: Optional[List[str]] = None,
        await_children: Optional[bool] = False,
        wrapper_class_instance: Optional[IsResource] = None,
        url_post_name: Optional[str] = None,
    ):
        self.alias_field = alias_field
        self.post_name = url_post_name
        self.resource_type = resource_type

        self.id = uuid
        self.parent = parent
        self.api_client = api_client
        self.children: Dict[Type[IsResource] : ResourceCache] = {}
        self.base_url = ""
        self.params = params if params else {}
        self.changed_params: Set[str] = set()
        self.params_to_serialize = params_to_serialize if params_to_serialize else []
        self.await_children = await_children
        self.wrapper_class_instance = wrapper_class_instance

        if parent:
            if not self.parent["id"]:
                log_error(
                    logger,
                    f"Parent resource {self.parent.resource_type} has no id",
                    ResourceIdMissing,
                )

            self.base_url = f'{self.parent["base_url"]}{self.parent.resource_type}/{self.parent["id"]}/'

            self.api_client: ApiClient = self.parent.api_client

        if not self.api_client:
            log_error(logger, "Api client has not been defined", ValueError)

    def get_alias_field_value(self, alias_field: str, params: Optional[dict]) -> str:
        """Returns the value of one of the fields used to create the alias
        :param alias_field: The name of the field
        :param params: The params to use to get the value of the field
        :return: The value of the field
        """
        if not params:
            params = self.params
        path = alias_field.split("/")
        alias = params[path[0]]
        for path_step in path[1:]:
            alias = alias[path_step]
        return alias

    def get_alias(self, params: Optional[dict] = None) -> Optional[Alias]:
        if self.alias_field:
            if isinstance(self.alias_field, tuple):
                return tuple(
                    self.get_alias_field_value(alias_field, params=params)
                    for alias_field in self.alias_field
                )
            if isinstance(self.alias_field, str):
                return self.get_alias_field_value(self.alias_field, params=params)
            log_error(
                logger,
                f"Alias field {self.alias_field} is not a string or a tuple",
                ValueError,
            )
        else:
            return None

    async def async_init(self):
        """
        Initializes the resource asynchronously. If the resource has no id, it will be created, otherwise it will
        fetch the resource and its children from the API.
        """
        if not self.id:
            self.id = await self.create()
        else:
            await self.get()

        if self.await_children:
            await asyncio.gather(*[cache.list() for cache in self.children.values()])
            await asyncio.gather(
                *[child[1] for c_l in self.children.values() for child in c_l]
            )

        return self.wrapper_class_instance

    async def get(self):
        """Fetches the resource from the API and updates the params."""
        endpoint = f"{self.base_url}{self.resource_type}/{self.id}"
        resource_dict = await self.api_client.query_element(
            method="GET",
            endpoint=endpoint,
        )

        for field, value in resource_dict.items():
            if field in self.params:
                self.params[field] = (
                    value
                    if field not in self.params_to_serialize
                    else json.loads(value if value else "{}")
                )

        self.changed_params = set()
        return resource_dict

    async def create(self) -> str:
        """Creates the resource in the API with the current params."""
        if self.post_name:
            endpoint = f"{self.base_url}{self.post_name}"
        else:
            endpoint = f"{self.base_url}{self.resource_type}"

        self.changed_params = set()

        params = {
            param: p_value
            for param, p_value in self.params.copy().items()
            if p_value is not None
        }

        for field in self.params_to_serialize:
            if field in params:
                try:
                    params[field] = json.dumps(params[field])
                except TypeError:
                    log_error(logger, f"Field {field} is not serializable", TypeError)

        for field, value in params.items():
            if is_dataclass(value):
                params[field] = asdict(value)

        params = {k: v for k, v in params.items() if v is not None}

        obj = await self.api_client.query_element(
            method="POST", endpoint=endpoint, **{"body_params": params}
        )

        for k, v in obj.items():
            if k in self.params:
                if v is None or k not in self.params_to_serialize:
                    self.params[k] = v
                else:
                    try:
                        self.params[k] = json.loads(v)
                    except TypeError:
                        log_error(logger, f"Field {k} is not deserializable", TypeError)

        return obj["id"]

    async def update(self) -> bool:
        """Updates the resource in the API with the current params."""
        endpoint = f"{self.base_url}{self.resource_type}/{self.id}"

        params = {k: v for k, v in self.params.items() if k in self.changed_params}
        self.changed_params = set()

        for field in self.params_to_serialize:
            if field in params:
                assert isinstance(params[field], (dict, list))
                params[field] = json.dumps(params[field])

        for field, value in params.items():
            if is_dataclass(value):
                params[field] = asdict(value)

        if params:
            await self.api_client.query_element(
                method="PATCH", endpoint=endpoint, **{"body_params": params}
            )
            return True
        logger.debug(f"No params to update for {self.resource_type} {str(self)}")
        return False

    async def delete(self):
        """Deletes the resource from the API."""
        endpoint = f"{self.base_url}{self.resource_type}/{self.id}"
        await self.api_client.query_element(
            method="DELETE",
            endpoint=endpoint,
        )

    async def create_children_batch(
        self,
        resource_class: Type[IsResource],
        children_params: List[Dict],
        unit: str,
        batch_size: int = 100,
    ):
        """Creates a batch of children of a given resource class. It doesn't return the created resources,
        And it doesn't save them in the cache.
        :param resource_class: The class of the resource to create.
        :param children_params: The parameters of the resources to create.
        :param unit: The unit of the progress bar.
        :param batch_size: The size of the batch to create.
        """
        if batch_size >= 1000:
            raise ValueError("batch_size must be less than 1000")

        endpoint = f"{self.base_url}{self.resource_type}/{self.id}/{resource_class.resource_type}/batch"

        log_level = logger.getEffectiveLevel()
        disable = log_level > logging.INFO or len(children_params) < 1000

        # TODO Seems to be converted in the api when returned? Solve in local server if that is the case
        if self.api_client.playground:
            # TODO maybe params_to_serialize should be a class attribute
            mock_resource = resource_class(parent=self.wrapper_class_instance)
            for child_params in children_params:
                for field in mock_resource.params_to_serialize:
                    if field in child_params:
                        child_params[field] = json.dumps(child_params[field])

        if not disable:
            logger.info("Uploading data")

        if not IN_BROWSER:
            progress_bar = tqdm(total=len(children_params), unit=unit, disable=disable)
        else:
            progress_bar = None

        query_tasks = []
        for chunk in range(0, len(children_params), batch_size):
            query_tasks.append(
                self.api_client.query_element(
                    method="POST",
                    endpoint=endpoint,
                    **{
                        "body_params": children_params[chunk : chunk + batch_size],
                        "progress_bar": (
                            progress_bar,
                            len(children_params[chunk : chunk + batch_size]),
                        ),
                    },
                )
            )
        await asyncio.gather(*query_tasks)

        logger.info("data uploaded") if not disable else None

    async def create_child(
        self, resource_class: Type[IsResource], **kwargs
    ) -> IsResource:
        """Creates a child of a given resource class.
        :param resource_class: The class of the resource to create.
        :param kwargs: The parameters of the resource to create.
        :return: The created resource.
        """
        res_cache: ResourceCache = self.children[resource_class]

        if "uuid" in kwargs:
            log_error(
                logger,
                f"Cannot create child with uuid {kwargs['uuid']}, use get instead",
                ValueError,
            )

        if "alias" in kwargs:
            if kwargs["alias"] is None:
                log_error(logger, "Cannot create child with alias None", ValueError)
            await res_cache.raise_if_alias_exists(kwargs["alias"])
            alias = kwargs["alias"]
            del kwargs["alias"]
            kwargs = dict(alias=alias, params=kwargs)
        else:
            resource = resource_class(parent=self.wrapper_class_instance)
            kwargs = dict(resource=resource, params=kwargs)

        return await res_cache.add(**kwargs)

    async def update_child(
        self,
        resource_class: Type[IsResource],
        uuid: Optional[str] = None,
        alias: Optional[Alias] = None,
        **params,
    ) -> bool:
        """Updates a child of a given resource class.
        :param resource_class: The class of the resource to update.
        :param uuid: The uuid of the resource to update.
        :param alias: The alias of the resource to update.
        :param params: The parameters of the resource to update.
        :return: True if the resource was updated, False otherwise.
        """
        return await self.children[resource_class].update(
            uuid=uuid, alias=alias, **params
        )

    async def get_children(
        self, resource_class: Type[IsResource], limit: Optional[int] = None
    ) -> List[IsResource]:
        """Gets all children of a given resource class.
        :param resource_class: The class of the resources to get.
        :param limit: The maximum number of resources to get.
        :return: A list of resources.
        """
        return await self.children[resource_class].list(limit)

    async def get_child(
        self,
        resource_class: Type[IsResource],
        uuid: Optional[str] = None,
        alias: Optional[Alias] = None,
        create_if_not_exists: bool = False,
    ) -> Optional[IsResource]:
        """Gets a child resource.
        :param resource_class: The class of the resource to get.
        :param uuid: The uuid of the resource to get.
        :param alias: The alias of the resource to get.
        :param create_if_not_exists: If True, creates the resource if it does not exist.
        """
        if not uuid and alias is None:
            log_error(logger, "Cannot get child without uuid or alias", ValueError)

        res_cache: ResourceCache = self.children[resource_class]

        child = await res_cache.get(uuid, alias)

        if not child and alias and create_if_not_exists:
            child = await self.create_child(resource_class, alias=alias)

        return child

    async def delete_child(
        self,
        resource_class: Type[IsResource],
        uuid: Optional[str] = None,
        alias: Optional[Alias] = None,
    ) -> bool:
        """Deletes a child resource.
        :param resource_class: The class of the resource to delete.
        :param uuid: The uuid of the resource to delete.
        :param alias: The alias of the resource to delete.
        """
        return await self.children[resource_class].delete(uuid, alias)

    def clear(self):
        """Clears the cache."""
        for resource_cache in self.children.values():
            resource_cache.clear()

    # @logging_before_and_after(logging_level=logger.debug)
    def cascade_to_dict(self) -> Dict[str, Any]:
        """Returns a dictionary representation of the resource and its children.
        :return: A dictionary representation of the resource and its children.
        """
        resource_dict = {
            "id": self.id,
        }
        resource_dict.update(self.params)

        for resource_class, children in self.children.items():
            resource_dict[resource_class.plural] = children.cascade_to_dict()

        return resource_dict

    def __str__(self):
        """Returns a string representation of the resource."""
        alias = self.get_alias()
        return self.id if not alias else alias

    def __eq__(self, other):
        """Returns True if the resource is equal to another resource."""
        return self.params == other.params


class Resource(ClassWithLogging, ABC):
    alias_field: Optional[AliasField] = None

    resource_type: Optional[str] = None
    plural: Optional[str] = None
    url_post_name: Optional[str] = None

    elastic_supported: bool = False

    def __init__(
        self,
        parent: Optional[IsResource] = None,
        uuid: Optional[str] = None,
        db_resource: Optional[dict[str, Any]] = None,
        api_client: Optional[ApiClient] = None,
        params: Optional[dict[str, Any]] = None,
        children: Optional[List[Type[IsResource]]] = None,
        check_params_before_creation: Optional[List[str]] = None,
        params_to_serialize: Optional[List[str]] = None,
        await_children: Optional[bool] = False,
    ):
        """Initializes a resource."""
        self._initializing = True
        if db_resource:
            uuid = db_resource["id"]
        else:
            db_resource = {}

        self._base_resource = BaseResource(
            api_client=api_client,
            uuid=uuid,
            parent=parent,
            params=params,
            resource_type=self.resource_type,
            alias_field=self.alias_field,
            params_to_serialize=params_to_serialize,
            await_children=await_children,
            wrapper_class_instance=self,
            url_post_name=self.url_post_name,
        )

        self._base_resource.children = (
            {res_class: ResourceCache(res_class, self) for res_class in children}
            if children
            else {}
        )

        self._check_params_before_creation = (
            check_params_before_creation if check_params_before_creation else []
        )
        self.api_client: ApiClient = self._base_resource.api_client
        self.parent = self._base_resource.parent
        self.params_to_serialize = self._base_resource.params_to_serialize
        self.children = self._base_resource.children

        for field, value in db_resource.items():
            if field in self:
                if field in self.params_to_serialize and value is not None:
                    value = json.loads(value)
                self[field] = value
        self._initializing = False

    def __await__(self):
        """Allows the resource to be awaited."""
        if not self._base_resource.id:
            for param_name in self._check_params_before_creation:
                if self._base_resource.params[param_name] is None:
                    log_error(
                        logger,
                        f"Mandatory parameter {param_name} has not been defined",
                        ValueError,
                    )

        return self._base_resource.async_init().__await__()

    async def get(self) -> Dict[str, Any]:
        return await self._base_resource.get()

    # @logging_before_and_after(logging_level=logger.debug)
    def cascade_to_dict(self) -> Dict[str, Any]:
        """Returns a dictionary representation of the resource and its children."""
        return self._base_resource.cascade_to_dict()

    def clear(self):
        """Clears the resource."""
        self._base_resource.clear()

    def set_params(self, **kwargs):
        """Sets the parameters of the resource."""
        for key, value in kwargs.items():
            if (
                isinstance(key, tuple)
                and len(key) == 2
                and isinstance(key[0], str)
                and isinstance(key[1], str)
            ):
                self[key[0]][key[1]] = value
            elif value is not None:
                self[key] = value

    def empty_changed_params(self):
        """Empties the changed_params set."""
        self._base_resource.changed_params = set()

    def get_alias_from_other_params(self, params: dict) -> Optional[str]:
        """Returns the alias of the resource based on the given parameters."""
        return self._base_resource.get_alias(params)

    def __getitem__(self, item):
        """Returns the value of the parameter with the given name."""
        if item == "id":
            return self._base_resource.id
        elif item == "base_url":
            return self._base_resource.base_url
        elif item == "alias":
            return self._base_resource.get_alias()
        elif item in self._base_resource.params:
            result = self._base_resource.params[item]
            if isinstance(result, (dict, list)) or is_dataclass(result):
                self._base_resource.changed_params.add(item)
            return result
        elif (
            isinstance(item, tuple)
            and len(item) == 2
            and isinstance(item[0], str)
            and isinstance(item[1], str)
            and item[0] in self._base_resource.params
        ):
            deeper_params = self._base_resource.params[item[0]]
            if item[1] not in deeper_params:
                log_error(
                    logger,
                    f"{self.__class__.__name__} {item[0]} does not contain {item[1]}",
                    KeyError,
                )
            return deeper_params[item[1]]
        else:
            log_error(
                logger,
                f"{self.__class__.__name__} parameters do not contain {item}",
                KeyError,
            )

    def __setitem__(self, key, value):
        """Sets the value of the parameter with the given name."""
        if key not in self._base_resource.params:
            log_error(
                logger,
                f"{self.__class__.__name__} parameters do not contain {key}",
                KeyError,
            )
        class_type = self._base_resource.params[key].__class__
        if not isinstance(value, class_type) and not isinstance(value, type(None)):
            if is_dataclass(class_type):
                value = class_type(
                    **{
                        k: v
                        for k, v in value.items()
                        if k in class_type.__dataclass_fields__
                    }
                )
            elif not isinstance(self._base_resource.params[key], type(None)):
                log_error(
                    logger,
                    f"Value {value} with type {value.__class__} for parameter {key} "
                    f"is not of type {self._base_resource.params[key].__class__}",
                    ValueError,
                )
        self._base_resource.params[key] = value if value is not None else class_type()
        if not self._initializing:
            self._base_resource.changed_params.add(key)

    def __contains__(self, item):
        """Returns whether the resource contains a parameter with the given name."""
        return item in self._base_resource.params

    def __iter__(self):
        """Returns an iterator over the parameters of the resource."""
        return iter(self._base_resource.params)

    def __str__(self):
        """Returns a string representation of the resource."""
        return self._base_resource.__str__()

    def __eq__(self, other: IsResource):
        """Returns whether the resource is equal to another resource."""
        return self._base_resource.__eq__(other._base_resource)



import datetime as dt

import asyncio
import json

from typing import List, Dict, Optional, Union, TYPE_CHECKING, TypedDict, Any

from shimoku.api.base_resource import Resource

if TYPE_CHECKING:
    from shimoku.api.resources.app import App

import logging
from shimoku.execution_logger import log_error

logger = logging.getLogger(__name__)


class Activity(Resource):
    """
    A class that contains all the information necessary to create and run an activity.
    """

    _module_logger = logger
    resource_type = "activity"
    alias_field = "name"
    plural = "activities"

    class ActivityParams(TypedDict):
        name: Optional[str]
        settings: dict
        activityTemplateWithMode: Optional[dict]
        universeApiKeyId: Optional[str]

    class Run(Resource):
        """
        Run of an activity, which is a single execution of the activity.
        """

        _module_logger = logger
        resource_type = "run"
        plural = "runs"

        class Log(Resource):
            """
            Log of an activity run.
            """

            _module_logger = logger
            resource_type = "log"
            plural = "logs"

            def __init__(
                self,
                parent: "Activity.Run",
                uuid: Optional[str] = None,
                db_resource: Optional[Dict] = None,
            ):
                params = dict(
                    dateTime=str(dt.datetime.now().isoformat())[:-3] + "Z",
                    message="message",
                    severity="INFO",
                    tags={},
                )

                super().__init__(
                    parent=parent,
                    uuid=uuid,
                    db_resource=db_resource,
                    check_params_before_creation=["message", "severity"],
                    params_to_serialize=["tags"],
                    params=params,
                )

        def __init__(
            self,
            parent: "Activity",
            uuid: Optional[str] = None,
            db_resource: Optional[Dict] = None,
        ):
            params = dict(
                settings={},
            )

            super().__init__(
                parent=parent,
                uuid=uuid,
                db_resource=db_resource,
                children=[Activity.Run.Log],
                params_to_serialize=["settings"],
                params=params,
            )

        async def create_log(
            self, message: str, severity: str, tags: Dict[str, str]
        ) -> "Activity.Run.Log":
            """
            Creates a log for the activity run.
            :param message: The message of the log.
            :param severity: The severity of the log.
            :param tags: The tags of the log.
            :return: The created log.
            """
            if tags and (
                not isinstance(tags, Dict)
                or any(
                    [
                        not isinstance(key, str) or not isinstance(value, str)
                        for key, value in tags.items()
                    ]
                )
            ):
                log_error(
                    logger, "The tags must be a dictionary of strings.", ValueError
                )

            return await self._base_resource.create_child(
                Activity.Run.Log, message=message, severity=severity, tags=tags
            )

        async def get_log(self, uuid: str) -> Optional["Activity.Run.Log"]:
            """
            Gets a log of the activity run.
            :param uuid: The uuid of the log.
            :return: The log.
            """
            return await self._base_resource.get_child(Activity.Run.Log, uuid)

        async def get_logs(self) -> List["Activity.Run.Log"]:
            """
            Gets all the logs of the activity run.
            :return: The logs.
            """
            logs = await self._base_resource.get_children(Activity.Run.Log)
            return sorted(logs, key=lambda log: log["dateTime"])

        async def trigger_webhook(self):
            """
            Triggers the webhook of the activity in a run.
            """
            endpoint = (
                self._base_resource.base_url
                + f"run/{self._base_resource.id}/triggerWebhook"
            )

            response = await self._base_resource.api_client.query_element(
                method="POST",
                endpoint=endpoint,
            )

            return response["STATUS"]

        def cascade_to_dict(self) -> Dict[str, Any]:
            """
            Returns the run as a dictionary.
            :return: The run as a dictionary.
            """
            run_dict = super().cascade_to_dict()
            run_dict["logs"] = sorted(run_dict["logs"], key=lambda log: log["dateTime"])
            return run_dict

    @staticmethod
    def _runs_ordering(run: "Activity.Run") -> dt.datetime:
        """Returns the datetime of the last log of the run. Used for sorting runs.
        :param run: The run.
        :return: The datetime of the last log of the run.
        """
        logs = run._base_resource.children[Activity.Run.Log]
        if len(logs) > 0:
            return max(
                [
                    dt.datetime.strptime(log["dateTime"], "%Y-%m-%dT%H:%M:%S.%fZ")
                    if log["dateTime"]
                    else dt.datetime.min
                    for log_id, log in logs
                ]
            )
        else:
            return dt.datetime.min

    @staticmethod
    async def sort_runs_by_log_time(runs: List["Activity.Run"]) -> List["Activity.Run"]:
        """Returns the runs ordered by the datetime of the last log of the run.
        :param runs: The runs.
        :return: The runs ordered by the datetime of the last log of the run.
        """
        await asyncio.gather(*[run.get_logs() for run in runs])
        return sorted(runs, key=Activity._runs_ordering)

    def __init__(
        self,
        parent: "App",
        uuid: Optional[str] = None,
        alias: Optional[str] = None,
        db_resource: Optional[Dict] = None,
    ):
        params: Activity.ActivityParams = dict(
            name=alias,
            settings={},
            activityTemplateWithMode=None,
            universeApiKeyId=None,
        )

        super().__init__(
            parent=parent,
            uuid=uuid,
            db_resource=db_resource,
            params=params,
            children=[Activity.Run],
            check_params_before_creation=["name"],
            params_to_serialize=["settings"],
        )

    async def delete(self):
        """Deletes the activity."""
        return await self._base_resource.delete()

    async def update(self):
        """Updates the activity."""
        return await self._base_resource.update()

    # Webhook methods
    async def create_webhook(self, url: str, method: str, headers: Dict[str, str]):
        """Creates a webhook for the activity.
        :param url: The url of the webhook.
        :param method: The method of the webhook.
        :param headers: The headers of the webhook.
        """
        if not headers:
            headers = {}
        if not isinstance(headers, Dict) or any(
            [
                not isinstance(key, str) or not isinstance(value, str)
                for key, value in headers.items()
            ]
        ):
            log_error(
                logger, "The headers must be a dictionary of strings.", ValueError
            )

        endpoint = (
            self._base_resource.base_url + f"activity/{self._base_resource.id}/webhook"
        )
        params = dict(
            url=url,
            method=method,
            headers=json.dumps(headers),
        )
        await self._base_resource.api_client.query_element(
            method="POST", endpoint=endpoint, **{"body_params": params}
        )

    # Run methods
    async def get_run(self, uuid: str) -> Optional["Activity.Run"]:
        """Gets a run of the activity.
        :param uuid: The uuid of the run.
        :return: The run.
        """
        return await self._base_resource.get_child(Activity.Run, uuid)

    async def get_runs(
        self, how_many_runs: Optional[int] = None
    ) -> List["Activity.Run"]:
        """Gets the last runs of the activity.
        :param how_many_runs: The number of runs to get.
        :return: The runs.
        """
        runs = await Activity.sort_runs_by_log_time(
            await self._base_resource.get_children(Activity.Run)
        )
        return runs[-how_many_runs:] if how_many_runs is not None else runs

    async def create_run(
        self, settings: Optional[Union[Dict, str]] = None
    ) -> "Activity.Run":
        """Creates a run of the activity.
        :param settings: The settings of the run or id of the run to copy the settings from.
        :return: The run.
        """
        if isinstance(settings, str):
            settings = (await self.get_run(settings))["settings"]
            assert isinstance(settings, dict)

        return await self._base_resource.create_child(Activity.Run, settings=settings)



from shimoku import Client
from .data.pie import pie_data


def pie(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "pie-test")

    shimoku_client.plt.pie(data="main data", names="date", values="x", order=0)
    shimoku_client.plt.pie(
        data=pie_data, names="name", values="value", order=1, rows_size=2, cols_size=12
    )



from shimoku import Client
from .data.indicator import indicator_data_variant_1, indicator_data_variant_2


def indicator(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "indicators")

    order = shimoku_client.plt.indicator(data=indicator_data_variant_1, order=0)
    order = shimoku_client.plt.indicator(
        data=indicator_data_variant_1 + indicator_data_variant_1[2:], order=order
    )

    shimoku_client.plt.indicator(
        data=indicator_data_variant_2, order=order, rows_size=1, cols_size=12
    )

    shimoku_client.set_menu_path("test", "indicators-vertical")
    order = shimoku_client.plt.indicator(
        data=indicator_data_variant_2 + indicator_data_variant_2,
        order=0,
        rows_size=1,
        cols_size=6,
        vertical="Title of the indicators",
    )
    order = shimoku_client.plt.indicator(
        data=indicator_data_variant_2,
        order=order,
        rows_size=2,
        cols_size=4,
        vertical=True,
    )
    order = shimoku_client.plt.indicator(
        data=indicator_data_variant_2[0],
        order=order,
        rows_size=8,
        cols_size=2,
        vertical="Title of the indicator",
    )
    shimoku_client.plt.indicator(
        data=indicator_data_variant_2[0],
        order=order,
        rows_size=8,
        cols_size=12,
        vertical=True,
    )



from shimoku import Client
from .data.html import html_data


def html(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "html-test")

    shimoku_client.plt.html(html=html_data, order=0)
    shimoku_client.plt.html(html=html_data, order=1, rows_size=2, cols_size=12)



from ...report import Report


class Indicator(Report):
    report_type = "INDICATOR"

    possible_values = dict(
        colors=["success", "warning", "error", "neutral", "caution"],
        align=["left", "center", "right"],
        variant=["default", "outlined", "contained", "topColor"],
    )

    default_properties = dict(
        **Report.default_properties,
        title=None,
        value=None,
        description=None,
        targetPath=None,
        color="neutral",
        align="right",
        variant="default",
        info=None,
        icon=None,
        bigIcon=None,
        hideIcons=False,
        backgroundImage=None,
    )



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame
from dataclasses import dataclass
import shimoku


class ComponentsLayerHeader:
    """
    This class is used to interact with the API at the component level.
    """

    def delete_component(
        self,
        uuid: str,
    ):
        """
        :param uuid: uuid of the component
        """
        pass

    def get_component(
        self,
        uuid: str,
    ) -> dict:
        """
        :param uuid: uuid of the component
        :return: component
        """
        pass

    def get_component_data_set_links(
        self,
        uuid: str,
    ) -> list:
        """
        :param uuid: uuid of the component
        :return: list of data set links
        """
        pass

    def get_components_in_sub_path(
        self,
        path: str,
    ) -> list:
        """
        :param path: path to the components
        :return: list of components
        """
        pass



form_groups = {
    f"form group {i}": [
        {
            "mapping": "country",
            "fieldName": f"Country {i}",
            "inputType": "select",
            "options": ["España", "Colombia"],
        },
        {
            "dependsOn": f"Country {i}",
            "mapping": "city",
            "fieldName": f"City {i}",
            "inputType": "select",
            "options": {
                "España": ["Madrid", "Barcelona"],
                "Colombia": ["Bogotá", "Medellin"],
            },
        },
    ]
    for i in range(4)
}

form_groups["Personal information"] = [
    {
        "mapping": "name",
        "fieldName": "name",
        "inputType": "text",
    },
    {
        "mapping": "surname",
        "fieldName": "surname",
        "inputType": "text",
    },
    {
        "mapping": "age",
        "fieldName": "age",
        "inputType": "number",
    },
    {
        "mapping": "tel",
        "fieldName": "phone",
        "inputType": "tel",
    },
    {
        "mapping": "gender",
        "fieldName": "Gender",
        "inputType": "radio",
        "options": ["Male", "Female", "No-binary", "Undefined"],
    },
    {
        "mapping": "email",
        "fieldName": "email",
        "inputType": "email",
    },
]

form_groups["Other data"] = [
    {
        "mapping": "skills",
        "fieldName": "Skills",
        "options": ["Backend", "Frontend", "UX/UI", "Api Builder", "DevOps"],
        "inputType": "checkbox",
    },
    {
        "mapping": "birthDay",
        "fieldName": "Birthday",
        "inputType": "date",
    },
    {
        "mapping": "onCompany",
        "fieldName": "time on Shimoku",
        "inputType": "dateRange",
    },
    {
        "mapping": "hobbies",
        "fieldName": "Hobbies",
        "inputType": "select",
        "options": ["Make Strong Api", "Sailing to Canarias", "Send Abracitos"],
    },
    {
        "mapping": "textField2",
        "fieldName": "Test Text",
        "inputType": "text",
    },
    {
        "mapping": "objectives",
        "fieldName": "Objetivos",
        "inputType": "multiSelect",
        "options": ["sleep", "close eyes", "awake"],
    },
]



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame
from dataclasses import dataclass
import shimoku


class BoardsLayerHeader:
    """
    This class is used to interact with the API at the board level.
    """

    def add_menu_path_in_board(
        self,
        menu_path_name: Optional[str] = None,
        menu_path_id: Optional[str] = None,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Add an app in a board
        :param menu_path_name: name of the board
        :param menu_path_id: UUID of the board
        :param name: name of the board
        :param uuid: UUID of the board
        """
        pass

    def create_board(
        self,
        name: str,
        order: Optional[int] = None,
        is_public: bool = False,
        is_disabled: bool = False,
        theme: Optional[dict] = None,
    ) -> dict:
        """
        Create a board
        :param name: name of the board
        :param order: order of the board
        :param is_public: is the board public
        :param is_disabled: is the board disabled
        :param theme: theme of the board
        :return: board metadata
        """
        pass

    def create_role(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        resource: Optional[str] = None,
        role_name: Optional[str] = None,
        permission: Optional[str] = None,
        target: Optional[str] = None,
    ) -> dict:
        """
        Create a new role at the resource level
        :param name: name of the resource
        :param uuid: uuid of the resource
        :param resource: resource of the role
        :param role_name: name of the role
        :param permission: permission of the role
        :param target: target of the role
        """
        pass

    def delete_board(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> bool:
        """
        Delete a board
        :param name: name of the board
        :param uuid: UUID of the board
        """
        pass

    def delete_role(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        role_id: Optional[str] = None,
        role_name: Optional[str] = None,
    ):
        """
        Delete the role at the resource level
        :param name: name of the resource
        :param uuid: uuid of the resource
        :param role_id: id of the role
        :param role_name: name of the role
        """
        pass

    def force_delete_board(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Force delete a board, this will delete the links between the board and the boards first, so that
        the board can always be deleted, but the apps will not be deleted
        :param name: name of the board
        :param uuid: UUID of the board
        """
        pass

    def get_board(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> Optional[dict]:
        """
        Get the board metadata
        :param name: name of the board
        :param uuid: UUID of the board
        :return: board metadata
        """
        pass

    def get_board_menu_path_ids(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """
        Get the list of the ids of the boards in the board
        :param name: name of the board
        :param uuid: UUID of the board
        :return: list of the ids of the boards in the board
        """
        pass

    def get_role(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        role_name: Optional[str] = None,
    ) -> list:
        """
        Get the role at the resource level
        :param name: name of the resource
        :param uuid: uuid of the resource
        :param role_name: name of the role
        """
        pass

    def get_roles(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """
        Get the roles at the resource level
        :param name: name of the resource
        :param uuid: uuid of the resource
        """
        pass

    def group_menu_paths(
        self,
        menu_path_names: Union[list[str], None, str] = None,
        menu_path_ids: Optional[list[str]] = None,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Add multiple boards in a board, if the board does not exist create it
        :param menu_path_names: list of board names
        :param menu_path_ids: list of board UUIDs
        :param name: name of the board
        :param uuid: UUID of the board
        """
        pass

    def is_menu_path_in_board(
        self,
        menu_path_name: Optional[str] = None,
        menu_path_id: Optional[str] = None,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> bool:
        """
        Check if an app is in a board
        :param menu_path_name: name of the board
        :param menu_path_id: UUID of the board
        :param name: name of the board
        :param uuid: UUID of the board
        :return: True if the app is in the board, False otherwise
        """
        pass

    def remove_all_menu_paths_from_board(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Delete all boards links of a board
        :param name: name of the board
        :param uuid: UUID of the board
        """
        pass

    def remove_menu_path_from_board(
        self,
        menu_path_name: Optional[str] = None,
        menu_path_id: Optional[str] = None,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Remove an app from a board
        :param menu_path_name: name of the board
        :param menu_path_id: UUID of the board
        :param name: name of the board
        :param uuid: UUID of the board
        """
        pass

    def update_board(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        new_name: Optional[str] = None,
        order: Optional[int] = None,
        is_public: Optional[bool] = None,
        is_disabled: Optional[bool] = None,
        theme: Optional[dict] = None,
    ) -> bool:
        """
        Update a board
        :param new_name: name of the board
        :param uuid: UUID of the board
        :param name: new name of the board
        :param order: new order of the board
        :param is_public: new public permission of the board
        :param is_disabled: new is_disabled of the board
        :param theme: new theme of the board
        """
        pass



""""""
import unittest
from os import getenv
from typing import Dict, List
from shimoku.exceptions import WorkspaceError, CacheError, APIError


from utils import initiate_shimoku

s = initiate_shimoku()

business_id: str = getenv("BUSINESS_ID")
mock: bool = getenv("MOCK") == "TRUE"
playground: bool = getenv("PLAYGROUND") == "TRUE"
BUSINESS_TEST_NAME = "auto-test"


class TestBusinesses(unittest.TestCase):
    def test_get_business(self):
        result = s.workspaces.get_workspace(uuid=business_id)
        assert result

    def test_get_fake_business(self):
        if not mock and not playground:
            with self.assertRaises(APIError):
                s.workspaces.get_workspace(
                    uuid="this is a test",
                )

    def test_create_and_delete_business(self):
        if playground:
            return

        if s.workspaces.get_workspace(name=BUSINESS_TEST_NAME):
            s.workspaces.delete_workspace(name=BUSINESS_TEST_NAME)

        business: Dict = s.workspaces.create_workspace(name=BUSINESS_TEST_NAME)
        business_id_: str = business["id"]

        assert len(business_id_) > 0

        business_from_db: Dict = s.workspaces.get_workspace(
            uuid=business_id_,
        )

        business_roles = s.workspaces.get_roles(name=BUSINESS_TEST_NAME)

        assert len(business_roles) == 4
        resources = ["DATA", "DATA_EXECUTION", "USER_MANAGEMENT", "BUSINESS_INFO"]
        for role in business_roles:
            assert role["role"] == "business_read"
            assert role["permission"] == "READ"
            assert role["resource"] in resources
            assert role["target"] == "GROUP"
            resources.remove(role["resource"])

        assert {
            k: v
            for k, v in business_from_db.items()
            if k
            in [
                "id",
                "name",
                "universe",
                "__typename",
            ]
        } == {
            k: v
            for k, v in business.items()
            if k
            in [
                "id",
                "name",
                "universe",
                "__typename",
            ]
        }
        del business_from_db

        s.workspaces.delete_workspace(uuid=business_id_)

        if not mock:
            with self.assertRaises(APIError):
                s.workspaces.get_workspace(
                    uuid=business_id_,
                )

    def test_theme_customization(self):
        s.workspaces.update_workspace(
            uuid=business_id,
            theme={},
        )
        business: Dict = s.workspaces.get_workspace(uuid=business_id)

        assert business["theme"] == {}

        theme = {
            "palette": {
                "primary": {
                    "main": "#2222FF",
                },
                "background": {
                    "default": "#23232F",
                },
            },
            "typography": {
                "h1": {
                    "fontSize": "60px",
                },
            },
            "custom": {
                "radius": {
                    "xs": "0px",
                    "s": "0px",
                    "m": "0px",
                    "l": "0px",
                    "xl": "0px",
                },
                "logo": "https://assets-global.website-files.com/619f9fe98661d321dc3beec7"
                "/621f5f09cd144b3dc06dc0fd_Logo-shimoku-white.svg",
            },
        }

        s.workspaces.update_workspace(uuid=business_id, theme=theme)

        business: Dict = s.workspaces.get_workspace(uuid=business_id)

        assert business["theme"] == theme

    def test_update_business(self):
        business_original: Dict = s.workspaces.get_workspace(uuid=business_id)

        business_name: str = business_original["name"]
        new_business_name: str = f"{business_name}_changed"

        s.workspaces.update_workspace(
            uuid=business_id,
            new_name=new_business_name,
        )

        new_business: Dict = s.workspaces.get_workspace(uuid=business_id)

        business_changed: Dict = s.workspaces.get_workspace(name=new_business_name)
        assert {
            k: v
            for k, v in business_changed.items()
            if k
            in [
                "id",
                "name",
                "universe",
                "__typename",
            ]
        } == {
            k: v
            for k, v in new_business.items()
            if k
            in [
                "id",
                "name",
                "universe",
                "__typename",
            ]
        }
        assert new_business_name == business_changed["name"]

        # Restore previous name
        s.workspaces.update_workspace(name=new_business_name, new_name=business_name)

        business_restored: Dict = s.workspaces.get_workspace(name=business_name)
        assert business_name == business_restored["name"]

    def test_get_business_apps(self):
        s.set_workspace(uuid=business_id)
        how_many_apps = len(s.workspaces.get_workspace_menu_paths(uuid=business_id))
        s.menu_paths.get_menu_path(name="business_app_test")
        apps: List[Dict] = s.workspaces.get_workspace_menu_paths(uuid=business_id)
        assert len(apps) == how_many_apps + 1
        s.menu_paths.delete_menu_path(name="business_app_test")
        apps: List[Dict] = s.workspaces.get_workspace_menu_paths(uuid=business_id)
        assert len(apps) == how_many_apps

    def test_cant_create_business_with_similar_name(self):
        if playground:
            return

        if s.workspaces.get_workspace(name=BUSINESS_TEST_NAME):
            s.workspaces.delete_workspace(name=BUSINESS_TEST_NAME)

        s.workspaces.create_workspace(name=BUSINESS_TEST_NAME)

        class SimilarBusinessCase(unittest.TestCase):
            def test_create_similar(self):
                with self.assertRaises(WorkspaceError):
                    s.workspaces.create_workspace(
                        name="auto-TEST",
                    )

            def test_update_similar(self):
                with self.assertRaises(WorkspaceError):
                    s.workspaces.update_workspace(
                        name=BUSINESS_TEST_NAME,
                        new_name="auto-TEST",
                    )

        sbc = SimilarBusinessCase()
        sbc.test_create_similar()
        sbc.test_update_similar()
        s.workspaces.delete_workspace(name=BUSINESS_TEST_NAME)

    def test_cant_get_with_id_and_name(self):
        if playground:
            return

        if s.workspaces.get_workspace(name=BUSINESS_TEST_NAME):
            s.workspaces.delete_workspace(name=BUSINESS_TEST_NAME)

        wid = s.workspaces.create_workspace(name=BUSINESS_TEST_NAME)["id"]

        with self.assertRaises(CacheError):
            s.set_workspace(uuid=wid, name=BUSINESS_TEST_NAME)



from typing import Optional
from shimoku.cli import CLIParser, CLIFuncParam
from shimoku.cli.cloud.cascade_get_resources import ResourceGetter, InitOptions
from shimoku.cli.utils import choose_from_menu

from shimoku.exceptions import ActionError
from shimoku.execution_logger import configure_logging, log_error

import logging

logger = logging.getLogger(__name__)


def add_update_parser(parser: Optional[CLIParser] = None):
    """
    Function to add the update parser to a parser
    :param parser: Parser to add the update parser to
    :return: Update parser
    """
    params = {"name": "update", "description": "Commands to update existing resources"}
    if parser:
        update_parser = parser.add_command(**params)
    else:
        update_parser = CLIParser(**params)

    common_args = [
        CLIFuncParam(
            name="local-port",
            arg_type=int,
            arg_help="Local port to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="environment",
            arg_type=str,
            arg_help="Environment to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="access-token",
            arg_type=str,
            arg_help="Access token to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="universe-id",
            arg_type=str,
            arg_help="Universe ID to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="workspace-id",
            arg_type=str,
            arg_help="Workspace ID to use",
            mandatory=False,
        ),
    ]

    module_functions = [action, workspace, menu_path, menu_order, board, boards_order]

    for func in module_functions:
        update_parser.decor_add_command(common_args=common_args)(func)

    return update_parser


if __name__ == "__main__":
    add_update_parser().parse_args()


async def action(
    action: str = CLIFuncParam(prompt=True),
    new_name: str = CLIFuncParam(mandatory=False),
    new_description: str = CLIFuncParam(mandatory=False),
    **kwargs,
):
    """Update the properties of an action
    :param action: Name or id of the action
    :param new_name: New name for the action
    :param new_description: New description for the action
    """
    resource_getter = ResourceGetter(InitOptions(action=action, **kwargs))
    actions_layer = await resource_getter.get_actions_layer()
    action = await resource_getter.get_action()
    if not action:
        log_error(
            logger,
            f"Action {action} not found",
            ActionError,
        )
    await actions_layer.update_action(
        uuid=action["id"],
        new_name=new_name,
        new_description=new_description,
    )


async def workspace(
    new_name: str = CLIFuncParam(mandatory=False),
    reset_theme: bool = CLIFuncParam(
        default=False, action="store_true", mandatory=False
    ),
    **kwargs,
):
    """Update the name or reset the theme of a workspace
    :param new_name: New name for the workspace
    :param reset_theme: Flag to reset the theme of the workspace
    """
    resource_getter = ResourceGetter(InitOptions(**kwargs))
    businesses_layer = await resource_getter.get_businesses_layer()
    business_id = resource_getter.init_opts.workspace_id
    if not reset_theme:
        await businesses_layer.update_workspace(business_id, new_name=new_name)
    else:
        await businesses_layer.update_workspace(
            business_id, new_name=new_name, theme={}
        )


async def menu_order(
    change_sub_paths: bool = CLIFuncParam(
        default=False, action="store_true", mandatory=False
    ),
    **kwargs,
):
    """Update the order of menu paths
    :param change_sub_paths: Flag to change the order of the sub paths
    """
    resource_getter = ResourceGetter(InitOptions(**kwargs))
    businesses_layer = await resource_getter.get_businesses_layer()
    business = await resource_getter.get_business()

    apps_layer = await resource_getter.get_apps_layer()
    app_names = [app["name"] for app in await business.get_apps()] + [""]

    apps_order = []
    print("Choose the order of the menu paths, press enter to finish")
    previous_logging_level = logging.root.level
    if previous_logging_level == logging.INFO:
        configure_logging("WARNING")
    while True:
        next_app = choose_from_menu(app_names, "Next menu path: ")
        if next_app:
            if change_sub_paths:
                sub_paths_order = []
                sub_paths = (
                    await apps_layer.get_menu_path_sub_paths(name=next_app)
                ) + [""]
                print(
                    f"Choose the order of the sub paths in the menu path({next_app}), press enter to finish"
                )
                while True:
                    next_sub_path = choose_from_menu(sub_paths, "Next sub path: ")
                    if next_sub_path:
                        sub_paths_order.append(next_sub_path)
                    else:
                        break
                    sub_paths.remove(next_sub_path)
                    if len(sub_paths) == 1:
                        break
                apps_order.append((next_app, sub_paths_order))
            else:
                apps_order.append(next_app)
        else:
            break
        app_names.remove(next_app)
        if len(app_names) == 1:
            break
    if previous_logging_level == logging.INFO:
        configure_logging("INFO")

    await businesses_layer.change_menu_order(
        uuid=resource_getter.init_opts.workspace_id, menu_order=apps_order
    )


async def boards_order(**kwargs):
    """Update the order of the boards in a workspace"""
    resource_getter = ResourceGetter(InitOptions(**kwargs))
    businesses_layer = await resource_getter.get_businesses_layer()
    business = await resource_getter.get_business()

    dashboard_names = [d["name"] for d in await business.get_dashboards()] + [""]
    dashboards_order = []
    print("Choose the order of the dashboards in the workspace, press enter to finish")
    while True:
        next_dashboard = choose_from_menu(dashboard_names, "Next board: ")
        if next_dashboard:
            dashboards_order.append(next_dashboard)
        else:
            break
        dashboard_names.remove(next_dashboard)
        if len(dashboard_names) == 1:
            break
    await businesses_layer.change_boards_order(
        uuid=resource_getter.init_opts.workspace_id, boards=dashboards_order
    )


async def board(
    is_public: Optional[str],
    is_disabled: Optional[str],
    board: str = CLIFuncParam(prompt=True),
    new_name: str = CLIFuncParam(mandatory=False),
    order: int = CLIFuncParam(mandatory=False),
    **kwargs,
):
    """
    Update the properties of a board
    :param is_public: (true/false) Set the board as public
    :param is_disabled: (true/false) Set the board as disabled
    :param board: Name or id of the board
    :param new_name: New name for the board
    :param order: New order for the board
    """
    if is_public is not None:
        is_public = is_public.lower() == "true"
    if is_disabled is not None:
        is_disabled = is_disabled.lower() == "true"
    resource_getter = ResourceGetter(InitOptions(board=board, **kwargs))
    dashboards_layer = await resource_getter.get_dashboards_layer()
    dashboard = await resource_getter.get_dashboard()
    await dashboards_layer.update_board(
        uuid=dashboard["id"],
        new_name=new_name,
        order=order,
        is_public=is_public,
        is_disabled=is_disabled,
    )


async def menu_path(
    hide_title: Optional[str],
    hide_path: Optional[str],
    show_breadcrumb: Optional[str],
    show_history_navigation: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    new_name: str = CLIFuncParam(mandatory=False),
    order: int = CLIFuncParam(mandatory=False),
    **kwargs,
):
    """
    Update the properties of a menu path
    :param menu_path: Name or id of the menu path
    :param new_name: New name for the menu path
    :param order: New order for the menu path
    :param hide_title: (true/false) Hide the title of the menu path
    :param hide_path: (true/false) Hide the path of the menu path
    :param show_breadcrumb: (true/false) Show the breadcrumb of the menu path
    :param show_history_navigation: (true/false) Show the history navigation of the menu path
    """
    if hide_title is not None:
        hide_title = hide_title.lower() == "true"
    if hide_path is not None:
        hide_path = hide_path.lower() == "true"
    if show_breadcrumb is not None:
        show_breadcrumb = show_breadcrumb.lower() == "true"
    if show_history_navigation is not None:
        show_history_navigation = show_history_navigation.lower() == "true"
    resource_getter = ResourceGetter(InitOptions(menu_path=menu_path, **kwargs))
    apps_layer = await resource_getter.get_apps_layer()
    app = await resource_getter.get_app()

    await apps_layer.update_menu_path(
        uuid=app["id"],
        new_name=new_name,
        order=order,
        hide_title=hide_title,
        hide_path=hide_path,
        show_breadcrumb=show_breadcrumb,
        show_history_navigation=show_history_navigation,
    )



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame
from dataclasses import dataclass
import shimoku


class ActivitiesLayerHeader:
    """
    This class is used to interact with the activities that are available in a menu path, through the API.
    """

    def create_activity(
        self,
        name: str,
        settings: Optional[dict] = None,
        template_id: Optional[str] = None,
        template_name_version: Optional[tuple[str, str]] = None,
        template_mode: str = "LIGHT",
        universe_api_key: str = "",
    ) -> dict:
        """
        Create an activity by its name and app id
        :param name: the name of the activity
        :param settings: the settings of the activity
        :param template_id: the template id of the activity
        :param template_name_version: the template name and version of the activity
        :param template_mode: the template mode of the activity
        :param universe_api_key: the universe api key of the activity
        :return: the dictionary representation of the activity
        """
        pass

    def create_run(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        settings: Union[dict, str] = None,
    ) -> dict:
        """
        Create a run for an activity by its name
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param settings: the settings of the run, or the id of the run to clone settings from
        :return: the run created as a dictionary
        """
        pass

    def create_run_log(
        self,
        run_id: str,
        message: str,
        severity: Optional[str] = "INFO",
        tags: Optional[dict[str, str]] = None,
        pretty_print: bool = False,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> dict:
        """
        Create a log for a run by its id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param run_id: the id of the run
        :param message: the message to log
        :param severity: the severity of the log
        :param tags: the tags of the log
        :param pretty_print: if True, the log is printed in a pretty format
        :return:
        """
        pass

    def create_webhook(
        self,
        url: str,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        method: str = "GET",
        headers: Optional[dict] = None,
    ):
        """
        Create a webhook by its name and app id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param url: the url of the webhook
        :param method: the method of the webhook
        :param headers: the headers of the webhook
        """
        pass

    def delete_activity(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        with_linked_to_templates: Optional[bool] = False,
    ):
        """
        Delete an activity by its name and app id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param with_linked_to_templates: if the activity is linked to a template, delete it anyway
        """
        pass

    def execute_activity(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        run_settings: Union[dict, str] = None,
    ) -> dict:
        """
        Execute an activity by its name or id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param run_settings: the settings of the run, or the id of the run to clone settings from
        :return: the run of the activity as a dictionary
        """
        pass

    def execute_run(
        self,
        run_id: str,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Execute a run by its id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param run_id: the id of the run
        :return:
        """
        pass

    def get_activities(
        self,
        pretty_print: bool = False,
        print_names: bool = False,
    ) -> list:
        """
        Get the list of activities in the app
        :param pretty_print: if True, the activities are printed in a pretty format
        :param print_names: if True and pretty_print is False, only the names of the activities are printed
        :return: a list of activities as dictionaries
        """
        pass

    def get_activity(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        pretty_print: bool = False,
        how_many_runs: Optional[int] = None,
    ) -> Optional[dict]:
        """
        Get an activity by its name
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param pretty_print: if True, the activity is printed in a pretty format
        :param how_many_runs: how many runs to get
        :return: the activity as a dictionary
        """
        pass

    def get_run_logs(
        self,
        run_id: str,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """
        Get the logs of a run by its id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param run_id: the id of the run
        :return: the logs of the run as a list of dictionaries
        """
        pass

    def get_run_settings(
        self,
        run_id: str,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> dict:
        """
        Get the settings of a run by its id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param run_id: the id of the run
        :return: the settings of the run as a dictionary
        """
        pass

    def update_activity(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        new_name: Optional[str] = None,
        settings: Optional[dict] = None,
    ):
        """
        Update an activity by its name or id
        :param name: the name of the activity
        :param uuid: the id of the activity
        :param new_name: the new name of the activity
        :param settings: the default settings of the activity
        :return: the updated activity as a dictionary
        """
        pass



from shimoku import Client
from .data.suburst import sunburst_data


def tree(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "tree-test")

    shimoku_client.plt.tree(data=sunburst_data, order=0)
    shimoku_client.plt.tree(
        data="tree_data", order=1, rows_size=4, cols_size=12, title="Tree", radial=True
    )
    shimoku_client.plt.tree(
        data="tree_data",
        order=2,
        rows_size=2,
        cols_size=12,
        title="Tree",
        vertical=True,
    )
    shimoku_client.plt.tree(
        data="tree_data",
        order=3,
        rows_size=4,
        cols_size=12,
        title="Tree",
        radial=True,
        vertical=True,
    )



from shimoku import Client
from .data.scatter_with_effect import scatter_source, effect_points


def scatter_with_effect(shimoku_client: Client):
    shimoku_client.set_menu_path("test-free-echarts", "scatter-with-effect-test")

    shimoku_client.plt.scatter_with_effect(
        data=scatter_source,
        x="x",
        y="y",
        order=0,
        x_axis_name="X axis",
        y_axis_name="Y axis",
        title="Scatter with effect",
        effect_points=effect_points,
    )



from typing import TYPE_CHECKING
from shimoku.api.utils import change_data_set_name_with_report
from shimoku.code_gen.business_code_gen.apps_code_gen.data_sets_code_gen.code_gen_from_data_sets import (
    code_gen_read_csv_from_data_set,
)
from shimoku.code_gen.utils_code_gen import code_gen_from_dict, code_gen_from_list

if TYPE_CHECKING:
    from shimoku.api.resources.report import Report
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import (
        AppCodeGen,
    )


def compact_labels_info(columns: list[dict], data_type_mapping: dict) -> dict:
    """Compact label info.
    :param columns: columns dict of the table
    :param data_type_mapping: mapping from field to data type
    :return: compacted label info
    """
    label_columns = {}
    for column in columns:
        if "chips" not in column:
            continue
        chips_dict = column["chips"]
        variant = chips_dict["variant"]
        mapping_from_field = data_type_mapping[column["field"]]
        entry_key = column["field"]
        chips_options = chips_dict["options"]
        if variant != "filled":
            entry_key = (entry_key, variant)
        entry_value = {}
        if mapping_from_field.startswith("intField"):
            chips_options = sorted(chips_options, key=lambda x: int(x["value"]))
            current_range = (0, chips_options[0]["value"])
            current_color = chips_options[0]["backgroundColor"]
            first_index = 0
            i = 1
            for chip in chips_options[1:]:
                if chip["backgroundColor"] != current_color:
                    if i == first_index + 1:
                        entry_value[current_range[1]] = current_color
                    else:
                        entry_value[current_range] = current_color
                    first_index = i
                    current_range = (current_range[1], chip["value"])
                    current_color = chip["backgroundColor"]
                else:
                    current_range = (current_range[0], chip["value"])
                i += 1
            if first_index == len(chips_options) - 1:
                entry_value[current_range[1]] = current_color
            else:
                entry_value[
                    (current_range[0], int(current_range[1]) + 1)
                ] = current_color
        else:
            entry_value = {
                chip["value"]: chip["backgroundColor"] for chip in chips_options
            }
        first_color = chips_options[0]["backgroundColor"]
        if all(v == first_color for v in entry_value.values()):
            entry_value = first_color
        label_columns[entry_key] = entry_value
    return label_columns


def get_column_options(columns: list[dict]) -> dict:
    """Get column options.
    :return: column options
    """
    column_options = {}
    for column in columns:
        not_wanted_keys = ["field", "headerName", "order", "type", "chips", "options"]
        single_column_options = {
            k: v for k, v in column.items() if k not in not_wanted_keys
        }
        if single_column_options:
            column_options[column["field"]] = single_column_options
    return column_options


async def code_gen_from_table(
    self: "AppCodeGen", report: "Report", properties: dict
) -> list[str]:
    """Generate code for a table report.
    :param report: report to generate code from
    :param properties: properties of the report
    :return: list of code lines
    """
    report_data_set: Report.ReportDataSet = (await report.get_report_data_sets())[0]
    data_set_id = report_data_set["dataSetId"]
    data_set = await self._app.get_data_set(data_set_id)
    data_arg = await code_gen_read_csv_from_data_set(
        data_set, change_data_set_name_with_report(data_set, report)
    )
    if data_arg is None:
        return ["pass"]
    if data_set_id in self._code_gen_tree.shared_data_sets:
        data_arg = f'"{data_set["name"]}",'
    table_params = []
    rev_mapping = {
        v: (k if data_set["columns"] else v)
        for k, v in properties["rows"]["mapping"].items()
    }
    mapping = properties["rows"]["mapping"]
    table_params.append(f"    columns={list(rev_mapping.values())},")
    if properties["pagination"]["pageSize"] != 10:
        table_params.append(f'    page_size={properties["pagination"]["pageSize"]},')
    if "pageSizeOptions" in properties["pagination"]:
        page_size_code_lines = code_gen_from_list(
            properties["pagination"]["pageSizeOptions"], 8
        )
        table_params.extend(
            [
                f"    page_size_options={page_size_code_lines[0][8:]}",
                *page_size_code_lines[1:],
            ]
        )
    if not properties["columnsButton"]:
        table_params.append("    columns_button=False,")
    if not properties["filtersButton"]:
        table_params.append("    filters=False,")
    if not properties["exportButton"]:
        table_params.append("    export_to_csv=False,")
    if not properties["search"]:
        table_params.append("    search=False,")
    if properties.get("sort"):
        sort_field = rev_mapping[properties["sort"]["field"]]
        sort_direction = properties["sort"]["direction"]
        table_params.append(f'    initial_sort_column="{sort_field}",')
        if sort_direction != "asc":
            table_params.append("    sort_descending=True,")
    column_options = get_column_options(properties["columns"])
    if column_options:
        column_options_code_lines = code_gen_from_dict(column_options, 8)
        table_params.extend(
            [
                f"    columns_options={column_options_code_lines[0][8:]}",
                *column_options_code_lines[1:],
            ]
        )

    categorical_columns = [
        col_dict["field"] if data_set["columns"] else mapping[col_dict["field"]]
        for col_dict in properties["columns"]
        if col_dict.get("type") == "singleSelect"
    ]
    if categorical_columns:
        table_params.append(f"    categorical_columns={categorical_columns},")
    label_columns = compact_labels_info(properties["columns"], mapping)
    if label_columns:
        label_columns_code_lines = code_gen_from_dict(label_columns, 4)
        table_params.extend(
            [
                f"    label_columns={label_columns_code_lines[0][4:]}",
                *label_columns_code_lines[1:],
            ]
        )
    return [
        "shimoku_client.plt.table(",
        f"    data={data_arg},",
        *self.code_gen_report_params(report.cascade_to_dict()),
        *table_params,
        ")",
    ]



from typing import Optional, Union, TYPE_CHECKING

from copy import deepcopy

from pandas import DataFrame

from shimoku.plt.EChart_definitions.default_echart_options import (
    get_common_echart_options,
    get_common_series_options,
)

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


async def top_bottom_area_charts(
    self: "PlotLayer",
    data: Union[str, DataFrame, list[dict]],
    order: Optional[int] = None,
    x: str = "x",
    top_names: Optional[list[str]] = None,
    bottom_names: Optional[list[str]] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    title: Optional[str] = None,
    x_axis_name: Optional[str] = None,
    top_axis_name: Optional[str] = None,
    bottom_axis_name: Optional[str] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
):
    if top_names is None:
        top_names = ["up"]

    if bottom_names is None:
        bottom_names = ["down"]

    common_options = get_common_echart_options()
    series_options = get_common_series_options()

    common_options.update(
        {
            "xAxis": [
                {
                    "name": x_axis_name if x_axis_name else "",
                    "type": "category",
                    "boundaryGap": True,
                    "axisLine": {"onZero": False},
                    "data": "#set_data#",
                }
            ],
            "yAxis": [
                {"name": top_axis_name if top_axis_name else "", "type": "value"},
                {
                    "name": bottom_axis_name if bottom_axis_name else "",
                    "nameLocation": "start",
                    "alignTicks": True,
                    "type": "value",
                    "inverse": True,
                },
            ],
        }
    )
    series_options.update(
        {
            "type": "line",
            "areaStyle": {"opacity": 0.5},
            "lineStyle": {"width": 1},
            "emphasis": {"focus": "series"},
            "markArea": {
                "silent": True,
                "itemStyle": {"opacity": 0.3},
            },
        }
    )
    series = []
    for name in bottom_names:
        aux_series_options = deepcopy(series_options)
        aux_series_options["name"] = name
        series.append(aux_series_options)

    for name in top_names:
        aux_series_options = deepcopy(series_options)
        aux_series_options["name"] = name
        aux_series_options["yAxisIndex"] = 1
        series.append(aux_series_options)

    await self._create_trend_chart(
        data_mapping_to_tuples=await self._choose_data(order, data),
        order=order,
        axes=x,
        values=bottom_names + top_names,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        title=title,
        x_axis_names=x_axis_name,
        y_axis_names=[bottom_axis_name, top_axis_name],
        echart_options=common_options,
        series_options=series,
        option_modifications=option_modifications,
        variant=variant,
    )


async def top_bottom_line_charts(
    self: "PlotLayer",
    data: Union[str, DataFrame, list[dict]],
    order: int,
    x: str,
    top_names: Optional[list[str]] = None,
    bottom_names: Optional[list[str]] = None,
    rows_size: int = 4,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    title: Optional[str] = None,
    x_axis_name: Optional[str] = None,
    top_axis_name: Optional[str] = None,
    bottom_axis_name: Optional[str] = None,
    option_modifications: Optional[dict] = None,
):
    if top_names is None:
        top_names = ["up"]

    if bottom_names is None:
        bottom_names = ["down"]

    common_options = get_common_echart_options()
    series_options = get_common_series_options()

    common_options.update(
        {
            "grid": [
                {"left": 60, "right": 50, "height": "35%"},
                {"left": 60, "right": 50, "top": "55%", "height": "35%"},
            ],
            "xAxis": [
                {
                    "name": x_axis_name if x_axis_name else "",
                    "type": "category",
                    "boundaryGap": False,
                    "axisLine": {"onZero": True},
                    "data": "#set_data#",
                    "label": {"show": False},
                },
                {
                    "gridIndex": 1,
                    "type": "category",
                    "boundaryGap": False,
                    "axisLine": {"onZero": True},
                    "position": "top",
                    "data": "#set_data#",
                },
            ],
            "yAxis": [
                {
                    "name": top_axis_name if top_axis_name else "",
                    "type": "value",
                    "nameLocation": "middle",
                    "nameGap": 40,
                },
                {
                    "gridIndex": 1,
                    "name": bottom_axis_name if bottom_axis_name else "",
                    "alignTicks": True,
                    "type": "value",
                    "inverse": True,
                    "nameLocation": "middle",
                    "nameGap": 40,
                },
            ],
        }
    )
    series_options.update(
        {
            "type": "line",
            "lineStyle": {"width": 1},
            "emphasis": {"focus": "series"},
            "markArea": {
                "silent": True,
                "itemStyle": {"opacity": 0.3},
            },
        }
    )
    series = []
    for name in bottom_names:
        aux_series_options = deepcopy(series_options)
        aux_series_options["name"] = name
        series.append(aux_series_options)

    for name in top_names:
        aux_series_options = deepcopy(series_options)
        aux_series_options["name"] = name
        aux_series_options["yAxisIndex"] = 1
        aux_series_options["xAxisIndex"] = 1
        series.append(aux_series_options)

    await self._create_trend_chart(
        data_mapping_to_tuples=await self._choose_data(order, data),
        order=order,
        axes=[x, x],
        values=bottom_names + top_names,
        option_modifications=option_modifications,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        title=title,
        x_axis_names=x_axis_name,
        y_axis_names=[bottom_axis_name, top_axis_name],
        echart_options=common_options,
        series_options=series,
    )



from shimoku import Client


class UsesClient:
    def __init__(self):
        print("This should work")

    def use_s(self, shimoku_client: Client):
        shimoku_client.set_workspace()
        shimoku_client.set_menu_path("Pyodide test")
        shimoku_client.plt.gauge_indicator(
            order=2,
            value=len(
                shimoku_client.menu_paths.get_menu_path_components(
                    name="test-free-echarts"
                )
            ),
            description="Trying to get the number of components",
            title="Pyodide test",
        )


def deep_no_use_s(shimoku_client: Client):
    pass


def no_use_s(shimoku_client: Client):
    deep_no_use_s(shimoku_client)


def deep_deep_use_s(shimoku_client: Client):
    shimoku_client.set_workspace()
    shimoku_client.set_menu_path("Pyodide test")
    shimoku_client.plt.gauge_indicator(
        order=1,
        value=len(
            shimoku_client.menu_paths.get_menu_path_components(name="test-free-echarts")
        ),
        description="Trying to get the number of components",
        title="Pyodide test",
    )


def use_s(shimoku_client: Client):
    def deep_use_s():
        deep_deep_use_s(shimoku_client)

    deep_use_s()


def action(shimoku_client: Client):
    shimoku_client._async_pool.ACTIONS_TEST = True
    print("This should work")
    no_use_s(shimoku_client)
    use_s(shimoku_client)

    uses_shimoku_client = UsesClient()
    uses_shimoku_client.use_s(shimoku_client)

    shimoku_client.run()



from typing import Optional, List

import tqdm
import subprocess
import os

from shimoku import Business
from shimoku.code_gen.business_code_gen.code_gen_from_business import BusinessCodeGen
from shimoku.code_gen.file_generator import CodeGenFileHandler

from shimoku.utils import create_function_name


async def generate_code(
    business_object: Business,
    business_id: str = "local",
    access_token: Optional[str] = None,
    universe_id: str = "local",
    environment: str = "production",
    output_path: str = "generated_code",
    menu_paths: Optional[list[str]] = None,
    use_black_formatter: bool = False,
    pbar: Optional[tqdm.tqdm] = None,
):
    """Main code generation function.
    :param business_object: business object to generate code for
    :param business_id: business id to use
    :param access_token: access token to use
    :param universe_id: universe id to use
    :param environment: environment to use
    :param output_path: output path to use
    :param menu_paths: list of menu paths to generate code for
    :param use_black_formatter: whether to use black formatter
    :param pbar: progress bar to use
    """
    await BusinessCodeGen(business_object, output_path).generate_code(
        business_id, menu_paths, pbar=pbar
    )
    business_name = create_function_name(business_object["name"])

    main_code_lines: List[str] = [
        "shimoku_client = Client(",
    ]
    if universe_id != "local":
        main_code_lines.extend(
            [f'    access_token="{access_token}",', f'    universe_id="{universe_id}",']
        )
    main_code_lines.extend(
        [
            f'    environment="{environment}",',
            "    async_execution=True,",
            '    verbosity="INFO"',
            ")",
            "",
            f"workspace_{business_name}(shimoku_client)",
            "",
            "shimoku_client.run()",
        ]
    )
    CodeGenFileHandler(output_path).generate_script_file(
        f"execute_workspace_{business_name}",
        [
            "from shimoku import Client",
            f"from {business_name}.workspace_{business_name} import workspace_{business_name}",
            "",
            "",
            "def main():",
            *[f"    {line}" for line in main_code_lines],
            "",
            "",
            'if __name__ == "__main__":',
            "    main()",
            "",
        ],
    )

    if use_black_formatter:
        # apply black formatting
        subprocess.run(["black", "-l", "80", os.path.join(output_path)])

    if pbar:
        pbar.update(1)



from typing import TYPE_CHECKING
from shimoku.code_gen.utils_code_gen import code_gen_from_dict

if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import (
        AppCodeGen,
    )
    from shimoku.api.resources.report import Report


async def code_gen_from_button_modal(
    self: "AppCodeGen", report: "Report", report_params: list[str]
) -> list[str]:
    modal_id = report["properties"]["events"]["onClick"][0]["params"]["modalId"]
    modal = await self._app.get_report(modal_id)
    path = modal["path"] + "_" if modal["path"] else ""
    name = modal["properties"]["hash"][len(path) :]
    return [
        "shimoku_client.plt.modal_button(",
        f'    modal="{name}",',
        *report_params,
        ")",
    ]


async def code_gen_from_button_activity(
    self: "AppCodeGen", report: "Report", report_params: list[str]
) -> list[str]:
    activity_id = report["properties"]["events"]["onClick"][0]["params"]["activityId"]
    activity = await self._app.get_activity(activity_id)
    return [
        "shimoku_client.plt.activity_button(",
        f'    activity_name="{activity["name"]}",',
        *report_params,
        ")",
    ]


async def code_gen_from_button_generic(
    self: "AppCodeGen", report: "Report", report_params: list[str]
) -> list[str]:
    events_code = code_gen_from_dict(report["properties"]["events"], 4)
    return [
        "shimoku_client.plt.button(",
        *report_params,
        f"    on_click_events={events_code[0][4:]}",
        *events_code[1:],
        ")",
    ]


async def code_gen_from_button(self: "AppCodeGen", report: "Report") -> list[str]:
    """Generate code for a button report.
    :param report: report to generate code from
    :return: list of code lines
    """
    report_params = [
        f'    label="{report["properties"]["text"]}",',
        f'    order={report["order"]},',
    ]
    if report["sizeColumns"] != 12:
        report_params.append(f'    cols_size={report["sizeColumns"]},')
    if report["sizeRows"] != 1:
        report_params.append(f'    rows_size={report["sizeRows"]},')
    if report["sizePadding"] != "0,0,0,0":
        report_params.append(f'    padding="{report["sizePadding"]}",')
    if report["properties"]["align"] != "stretch":
        report_params.append(f'    align="{report["properties"]["align"]}",')

    if report["properties"]["events"]["onClick"][0]["action"] == "openModal":
        return await code_gen_from_button_modal(self, report, report_params)
    elif report["properties"]["events"]["onClick"][0]["action"] == "openActivity":
        return await code_gen_from_button_activity(self, report, report_params)
    else:
        return await code_gen_from_button_generic(self, report, report_params)



from shimoku import Client
from .data.sankey import sankey_data


def sankey(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "sankey-test")

    shimoku_client.plt.sankey(
        data=sankey_data, sources="source", targets="target", values="value", order=0
    )



from typing import Optional
from shimoku.cli import CLIParser, CLIFuncParam
from shimoku.cli.cloud.cascade_get_resources import InitOptions, ResourceGetter

import asyncio

import logging

logger = logging.getLogger(__name__)


def add_delete_parser(parser: Optional[CLIParser] = None):
    """
    Function to add the delete parser to a parser
    :param parser: Parser to add the delete parser to
    :return: Delete parser
    """
    params = {"name": "delete", "description": "Commands to create resources"}
    if parser:
        delete_parser = parser.add_command(**params)
    else:
        delete_parser = CLIParser(**params)

    common_args = [
        CLIFuncParam(
            name="local-port",
            arg_type=int,
            arg_help="Local port to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="environment",
            arg_type=str,
            arg_help="Environment to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="access-token",
            arg_type=str,
            arg_help="Access token to use",
            mandatory=False,
        ),
        CLIFuncParam(
            name="universe-id",
            arg_type=str,
            arg_help="Universe ID to use",
            mandatory=False,
        ),
    ]

    module_functions = [
        universe_api_key,
        action,
        workspace,
        workspace_contents,
        board,
        boards,
        board_menu_path_link,
        board_menu_path_links,
        menu_path,
        menu_paths,
        component,
        components,
        data_set,
        data_sets,
        file,
        files,
        ai_input_file,
        ai_model,
        role,
        invitation,
        workspace_user,
    ]

    for func in module_functions:
        delete_parser.decor_add_command(common_args=common_args)(func)

    return delete_parser


async def action(
    action: str = CLIFuncParam(prompt=True),
    **kwargs,
):
    """
    Delete an action
    :param action: Action id or name to delete
    """
    resource_getter = ResourceGetter(InitOptions(action=action, **kwargs))
    actions_layer = await resource_getter.get_actions_layer()
    action_obj = await resource_getter.get_action()
    if not action_obj:
        print(f"Action {action} not found")
        return
    await actions_layer.delete_action(uuid=action_obj["id"])


async def universe_api_key(uuid: str = CLIFuncParam(prompt=True), **kwargs):
    """Delete a universe api key
    :param uuid: UUID of the universe api key
    """
    resource_getter = ResourceGetter(InitOptions(**kwargs))
    universes_layer = await resource_getter.get_universes_layer()
    await universes_layer.delete_universe_api_key(
        uuid=kwargs["uuid"], api_key_uuid=uuid
    )


async def workspace(workspace_id: str = CLIFuncParam(prompt=True), **kwargs):
    """Delete the workspace
    :param workspace_id: UUID of the workspace to use
    """
    confirmation = input("Are you sure you want to delete the workspace? (y/n): ")
    if confirmation.lower() != "y" and confirmation.lower() != "yes":
        logger.info("Cancelling delete workspace")
        return
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, **kwargs))
    businesses_layer = await resource_getter.get_businesses_layer()
    await businesses_layer.delete_workspace(workspace_id)


async def workspace_contents(
    workspace_id: Optional[str],
    force: bool = CLIFuncParam(default=False, action="store_true", mandatory=False),
    **kwargs,
):
    """Delete all contents in the workspace
    :param workspace_id: UUID of the workspace to use
    :param force: Flag to not ask for confirmation
    """
    if not force:
        confirmation = input(
            "Are you sure you want to delete all contents in the workspace?"
            "\nthis action will delete all files, data sets and activities in the workspace"
            "\nmake sure to not lose any sensitive data. (y/n): "
        )
        if confirmation.lower() != "y" and confirmation.lower() != "yes":
            logger.info("Cancelling delete all workspace contents")
            return
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, **kwargs))
    businesses_layer = await resource_getter.get_businesses_layer()
    app_layer = await resource_getter.get_apps_layer()
    business = await resource_getter.get_business()
    await asyncio.gather(
        *[
            app_layer.delete_all_menu_path_activities(app["id"])
            for app in await business.get_apps()
        ]
    )
    await businesses_layer.delete_all_workspace_menu_paths(business["id"])
    await businesses_layer.delete_all_workspace_boards(business["id"])


async def board(
    workspace_id: Optional[str],
    board: str = CLIFuncParam(prompt=True),
    force: bool = CLIFuncParam(default=False, action="store_true", mandatory=False),
    **kwargs,
):
    """Delete the board in the workspace
    :param workspace_id: UUID of the workspace to use
    :param board: Board id or name to delete
    :param force: Flag to force delete the board
    """
    resource_getter = ResourceGetter(
        InitOptions(workspace_id=workspace_id, board=board, **kwargs)
    )
    dashboards_layer = await resource_getter.get_dashboards_layer()
    dashboard = await resource_getter.get_dashboard()
    if not dashboard:
        print(f"Board {board} not found")
        return
    if force:
        await dashboards_layer.force_delete_board(dashboard["id"])
    else:
        await dashboards_layer.delete_board(dashboard["id"])


async def boards(
    workspace_id: Optional[str],
    force: bool = CLIFuncParam(default=False, action="store_true", mandatory=False),
    **kwargs,
):
    """Delete all boards in the workspace
    :param workspace_id: UUID of the workspace to use
    :param force: Flag to force delete the boards
    """
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, **kwargs))
    businesses_layer = await resource_getter.get_businesses_layer()
    business = await resource_getter.get_business()
    await businesses_layer.delete_all_workspace_boards(business["id"], force=force)


async def board_menu_path_link(
    workspace_id: Optional[str],
    board: str = CLIFuncParam(prompt=True),
    menu_path: str = CLIFuncParam(prompt=True),
    **kwargs,
):
    """Delete the board menu path link
    :param workspace_id: UUID of the workspace to use
    :param board: Board id or name to use
    :param menu_path: Menu path id or name to delete
    """
    resource_getter = ResourceGetter(
        InitOptions(
            board=board, menu_path=menu_path, workspace_id=workspace_id, **kwargs
        )
    )
    dashboards_layer = await resource_getter.get_dashboards_layer()
    dashboard = await resource_getter.get_dashboard()
    app = await resource_getter.get_app()
    if not dashboard:
        print(f"Board {board} not found")
        return
    if not app:
        print(f"Menu path {menu_path} not found")
        return
    await dashboards_layer.remove_menu_path_from_board(
        menu_path_id=app["id"], uuid=dashboard["id"]
    )


async def board_menu_path_links(
    workspace_id: Optional[str], board: str = CLIFuncParam(prompt=True), **kwargs
):
    """Delete all board menu path links
    :param workspace_id: UUID of the workspace to use
    :param board: Board id or name to use
    """
    resource_getter = ResourceGetter(
        InitOptions(board=board, workspace_id=workspace_id, **kwargs)
    )
    dashboards_layer = await resource_getter.get_dashboards_layer()
    dashboard = await resource_getter.get_dashboard()
    if not dashboard:
        print(f"Board {board} not found")
        return
    await dashboards_layer.remove_all_menu_paths_from_board(dashboard["id"])


async def menu_path(
    workspace_id: Optional[str], menu_path: str = CLIFuncParam(prompt=True), **kwargs
):
    """Delete the menu path in the workspace
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to delete
    """
    resource_getter = ResourceGetter(
        InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs)
    )
    apps_layer = await resource_getter.get_apps_layer()
    app = await resource_getter.get_app()
    if not app:
        print(f"Menu path {menu_path} not found")
        return
    await apps_layer.delete_menu_path(uuid=app["id"])


async def menu_paths(workspace_id: Optional[str], **kwargs):
    """Delete all menu paths in the workspace
    :param workspace_id: UUID of the workspace to use
    """
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, **kwargs))
    businesses_layer = await resource_getter.get_businesses_layer()
    business = await resource_getter.get_business()
    await businesses_layer.delete_all_workspace_menu_paths(business["id"])


async def component(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    component_id: str = CLIFuncParam(prompt=True),
    **kwargs,
):
    """Delete the component from a menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param component_id: Component id or name to delete
    """
    resource_getter = ResourceGetter(
        InitOptions(
            workspace_id=workspace_id,
            menu_path=menu_path,
            component_id=component_id,
            **kwargs,
        )
    )
    reports_layer = await resource_getter.get_reports_layer()
    report = await resource_getter.get_report()
    if not report:
        print(f"Component {component_id} not found")
        return
    await reports_layer.delete_component(uuid=report["id"])


async def components(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    sub_path: str = CLIFuncParam(mandatory=False),
    **kwargs,
):
    """Delete all components from a menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param sub_path: Sub path to clear
    """
    resource_getter = ResourceGetter(
        InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs)
    )
    apps_layer = await resource_getter.get_apps_layer()
    app = await resource_getter.get_app()
    await apps_layer.delete_all_menu_path_components(uuid=app["id"], sub_path=sub_path)


async def data_set(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    data_set: str = CLIFuncParam(prompt=True),
    **kwargs,
):
    """Delete the data set from a menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param data_set: Data set id or name to delete
    """
    resource_getter = ResourceGetter(
        InitOptions(
            workspace_id=workspace_id, menu_path=menu_path, data_set=data_set, **kwargs
        )
    )
    data_sets_layer = await resource_getter.get_data_sets_layer()
    data_set_obj = await resource_getter.get_data_set()
    if not data_set_obj:
        print(f"Data set {data_set} not found")
        return
    await data_sets_layer.delete_data_set(uuid=data_set_obj["id"])


async def data_sets(
    workspace_id: Optional[str], menu_path: str = CLIFuncParam(prompt=True), **kwargs
):
    """Delete all data sets from a menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    """
    resource_getter = ResourceGetter(
        InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs)
    )
    apps_layer = await resource_getter.get_apps_layer()
    app = await resource_getter.get_app()
    await apps_layer.delete_all_menu_path_data_sets(uuid=app["id"])


async def activity(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    activity: str = CLIFuncParam(prompt=True),
    with_linked_to_templates: bool = CLIFuncParam(
        default=False, action="store_true", mandatory=False
    ),
    **kwargs,
):
    """Delete the activity from a menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param activity: Activity id or name to delete
    :param with_linked_to_templates: Flag to delete the activity even if it's linked to a template
    """
    resource_getter = ResourceGetter(
        InitOptions(
            workspace_id=workspace_id, menu_path=menu_path, activity=activity, **kwargs
        )
    )
    activities_layer = await resource_getter.get_activities_layer()
    activity = await resource_getter.get_activity()
    if not activity:
        print(f"Activity {activity} not found")
        return
    await activities_layer.delete_activity(
        uuid=activity["id"], with_linked_to_templates=with_linked_to_templates
    )


async def activities(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    with_linked_to_templates: bool = CLIFuncParam(
        default=False, action="store_true", mandatory=False
    ),
    **kwargs,
):
    """Delete all activities from a menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param with_linked_to_templates: Flag to delete the activities even if they were linked to templates
    """
    resource_getter = ResourceGetter(
        InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs)
    )
    apps_layer = await resource_getter.get_apps_layer()
    app = await resource_getter.get_app()
    await apps_layer.delete_all_menu_path_activities(
        uuid=app["id"], with_linked_to_templates=with_linked_to_templates
    )


async def file(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    file: str = CLIFuncParam(prompt=True),
    with_shimoku_generated: bool = CLIFuncParam(
        default=False, action="store_true", mandatory=False
    ),
    **kwargs,
):
    """Delete the file from a menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param file: File id or name to delete
    :param with_shimoku_generated: Flag to delete the file even if it was generated by shimoku
    """
    resource_getter = ResourceGetter(
        InitOptions(workspace_id=workspace_id, menu_path=menu_path, file=file, **kwargs)
    )
    files_layer = await resource_getter.get_files_layer()
    file = await resource_getter.get_file()
    if not file:
        print(f"File {file} not found")
        return
    print(file["id"])
    await files_layer.delete_file(
        uuid=file["id"], with_shimoku_generated=with_shimoku_generated
    )


async def files(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    with_shimoku_generated: bool = CLIFuncParam(
        default=False, action="store_true", mandatory=False
    ),
    **kwargs,
):
    """Delete all files from a menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param with_shimoku_generated: Flag to delete the files even if they were generated by shimoku
    """
    resource_getter = ResourceGetter(
        InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs)
    )
    apps_layer = await resource_getter.get_apps_layer()
    app = await resource_getter.get_app()
    await apps_layer.delete_all_menu_path_files(
        uuid=app["id"], with_shimoku_generated=with_shimoku_generated
    )


async def ai_input_file(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    name: str = CLIFuncParam(prompt=True),
    **kwargs,
):
    """Delete the AI input file from a menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param name: AI input file name to delete
    """
    resource_getter = ResourceGetter(
        InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs)
    )
    ai_layer = await resource_getter.get_ai_layer()
    await ai_layer.delete_input_file(file_name=name)


async def ai_model(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    name: str = CLIFuncParam(prompt=True),
    **kwargs,
):
    """Delete the AI model from a menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param name: AI model name to delete
    """
    resource_getter = ResourceGetter(
        InitOptions(workspace_id=workspace_id, menu_path=menu_path, **kwargs)
    )
    ai_layer = await resource_getter.get_ai_layer()
    await ai_layer.delete_model(model_name=name)


async def role(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(group="menu_path_or_board", mandatory=False),
    board: str = CLIFuncParam(group="menu_path_or_board", mandatory=False),
    role: str = CLIFuncParam(prompt=True),
    **kwargs,
):
    """Get a role from a specific level, leave menu path and board empty to get a workspace role
    and specify only one to get a menu path or board role respectively. Don't specify both.
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path name or id to use
    :param board: Board name or id to use
    :param role: Role name or id to get
    """
    resource_getter = ResourceGetter(
        InitOptions(
            workspace_id=workspace_id,
            menu_path=menu_path,
            board=board,
            role=role,
            **kwargs,
        )
    )
    role_obj = await resource_getter.get_role_dict()

    if not role_obj:
        logger.info(f"Role {role} not found")
        return
    if menu_path:
        layer = await resource_getter.get_apps_layer()
        resource_id = (await resource_getter.get_app())["id"]
    elif board:
        layer = await resource_getter.get_dashboards_layer()
        resource_id = (await resource_getter.get_dashboard())["id"]
    else:
        layer = await resource_getter.get_businesses_layer()
        resource_id = (await resource_getter.get_business())["id"]

    await layer.delete_role(uuid=resource_id, role_name=role_obj["role"])


async def invitation(
    workspace_id: Optional[str],
    email: str = CLIFuncParam(prompt=True),
    **kwargs,
):
    """Delete an invitation
    :param workspace_id: UUID of the workspace to use
    :param email: Email of the invitation to delete
    """
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, **kwargs))
    businesses_layer = await resource_getter.get_businesses_layer()
    business = await resource_getter.get_business()
    await businesses_layer.delete_pending_invitation(email=email, uuid=business["id"])


async def workspace_user(
    workspace_id: Optional[str],
    email: str = CLIFuncParam(prompt=True),
    **kwargs,
):
    """Delete a workspace user
    :param workspace_id: UUID of the workspace to use
    :param email: Email of the user to delete
    """
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, **kwargs))
    businesses_layer = await resource_getter.get_businesses_layer()
    business = await resource_getter.get_business()
    await businesses_layer.remove_user_from_workspace(email=email, uuid=business["id"])


if __name__ == "__main__":
    add_delete_parser().parse_args()



import unittest
from test_plot_api import main
from utils import initiate_shimoku
from shimoku import Client
from shimoku.exceptions import BoardError
import tempfile
import subprocess
import shutil
import json
import datetime as dt
from deepdiff import DeepDiff
from os import getenv

mock = getenv("MOCK") == "TRUE"


def get_report_data_set_info(
    r_hash: str, report_data_set: dict, data_sets_mapping: dict[str, str]
) -> dict:
    report_data_set.pop("id")
    report_id = report_data_set.pop("reportId")
    report_data_set["dataSetId"] = data_sets_mapping[
        report_data_set["dataSetId"]
    ].replace(report_id, r_hash)
    return report_data_set


def get_report_data_set_info_dict(
    r_hash: str,
    _dict: dict,
    report_data_sets: dict[str, dict],
    data_sets_mapping: dict[str, str],
):
    for k, v in _dict.items():
        if isinstance(v, dict):
            get_report_data_set_info_dict(
                r_hash, v, report_data_sets, data_sets_mapping
            )
        elif isinstance(v, list):
            get_report_data_set_info_list(
                r_hash, v, report_data_sets, data_sets_mapping
            )
        elif isinstance(v, str) and v.startswith("#{") and v.endswith("}"):
            uuid = v[2:-1]
            _dict[k] = get_report_data_set_info(
                r_hash, report_data_sets[uuid], data_sets_mapping
            )


def get_report_data_set_info_list(
    r_hash: str,
    _list: list,
    report_data_sets: dict[str, dict],
    data_sets_mapping: dict[str, str],
):
    for i, v in enumerate(_list):
        if isinstance(v, dict):
            get_report_data_set_info_dict(
                r_hash, v, report_data_sets, data_sets_mapping
            )
        elif isinstance(v, list):
            get_report_data_set_info_list(
                r_hash, v, report_data_sets, data_sets_mapping
            )
        elif isinstance(v, str) and v.startswith("#{") and v.endswith("}"):
            uuid = v[2:-1]
            _list[i] = get_report_data_set_info(
                r_hash, report_data_sets[uuid], data_sets_mapping
            )


def clear_workspace(shimoku_client):
    for menu_path in shimoku_client.workspaces.get_workspace_menu_paths(
        shimoku_client.workspace_id
    ):
        shimoku_client.menu_paths.delete_all_menu_path_activities(uuid=menu_path["id"])
    shimoku_client.workspaces.delete_all_workspace_menu_paths(
        shimoku_client.workspace_id
    )
    shimoku_client.workspaces.delete_all_workspace_boards(shimoku_client.workspace_id)


def change_ids_in_tabs_or_modal(component, components_by_id):
    if "tabs" in component["properties"]:
        for tab_name, tab_dict in component["properties"]["tabs"].items():
            component["properties"]["tabs"][tab_name]["reportIds"] = sorted(
                [
                    components_by_id[r_id]["properties"]["hash"]
                    for r_id in tab_dict["reportIds"]
                ]
            )
    if "reportIds" in component["properties"]:
        component["properties"]["reportIds"] = sorted(
            [
                components_by_id[r_id]["properties"]["hash"]
                for r_id in component["properties"]["reportIds"]
            ]
        )


def handle_events(events, components_by_id):
    on_click_events = events.get("onClick", [])
    for event in on_click_events:
        if "modalId" in event["params"]:
            event["params"]["modalId"] = components_by_id[event["params"]["modalId"]][
                "properties"
            ]["hash"]


def handle_bentobox(component, current_bentobox):
    if not component.get("bentobox"):
        current_bentobox["id_to_replace"] = ""
        return
    bentobox_id = component["bentobox"]["bentoboxId"]

    if bentobox_id != current_bentobox["id_to_replace"]:
        current_bentobox["id_to_replace"] = bentobox_id
        current_bentobox["bentoboxId"] = f'_{component["order"]}'
        current_bentobox["bentoboxOrder"] = component["order"]

    component["bentobox"]["bentoboxId"] = current_bentobox["bentoboxId"]
    component["bentobox"]["bentoboxOrder"] = current_bentobox["bentoboxOrder"]


def handle_component(
    shimoku_client,
    component,
    mpath_components_by_id,
    data_sets_mapping,
    current_bentobox,
):
    data_set_links = shimoku_client.components.get_component_data_set_links(
        component["id"]
    )
    get_report_data_set_info_dict(
        component["properties"]["hash"],
        component["properties"],
        {rds["id"]: rds for rds in data_set_links},
        data_sets_mapping,
    )
    handle_bentobox(component, current_bentobox)
    if "events" in component["properties"]:
        handle_events(component["properties"]["events"], mpath_components_by_id)
    if component["reportType"] in ["MODAL", "TABS"]:
        change_ids_in_tabs_or_modal(component, mpath_components_by_id)
    elif component["reportType"] == "FILTERDATASET":
        fds_mapping = component["properties"]["mapping"]
        fds_mapping[0]["id"] = data_sets_mapping[fds_mapping[0]["id"]]


def handle_menu_path(
    shimoku_client, menu_path, components, data_sets_mapping, data_by_data_set
):
    shimoku_client.set_menu_path(menu_path["name"], dont_add_to_dashboard=True)
    mpath_components = sorted(
        shimoku_client.menu_paths.get_menu_path_components(menu_path["id"]),
        key=lambda x: x["properties"]["hash"],
    )
    components.extend(mpath_components)
    components_hash_by_id = {c["id"]: c["properties"]["hash"] for c in mpath_components}
    data_sets = shimoku_client.menu_paths.get_menu_path_data_sets(menu_path["id"])
    for data_set in data_sets:
        for comp_id in components_hash_by_id.keys():
            if comp_id in data_set["name"]:
                data_set["name"] = data_set["name"].replace(
                    comp_id, components_hash_by_id[comp_id]
                )
        data_sets_mapping[data_set["id"]] = data_set["name"]

    for data_set in sorted(data_sets, key=lambda x: x["name"]):
        data_by_data_set[data_set["name"]] = shimoku_client.data.get_data_from_data_set(
            data_set["id"]
        )

    mpath_components_by_id = {c["id"]: c for c in mpath_components}
    current_bentobox = {"id_to_replace": "", "bentoboxId": "", "bentoboxOrder": ""}
    for component in mpath_components:
        handle_component(
            shimoku_client,
            component,
            mpath_components_by_id,
            data_sets_mapping,
            current_bentobox,
        )


def get_diff_percentage(items1, items2):
    return sum(DeepDiff(c1, c2) != {} for c1, c2 in zip(items1, items2)) / len(items1)


def get_workspace_contents(shimoku_client):
    boards = sorted(
        shimoku_client.workspaces.get_workspace_boards(shimoku_client.workspace_id),
        key=lambda x: x["order"],
    )
    menu_paths = sorted(
        shimoku_client.workspaces.get_workspace_menu_paths(shimoku_client.workspace_id),
        key=lambda x: x["order"],
    )
    components = []
    data_sets_mapping = {}
    data_by_data_set = {}
    for menu_path in menu_paths:
        handle_menu_path(
            shimoku_client, menu_path, components, data_sets_mapping, data_by_data_set
        )
        shimoku_client.pop_out_of_menu_path()
    _all = [
        *boards,
        *menu_paths,
        *components,
        *[datum for data in data_by_data_set.values() for datum in data],
    ]
    for element in _all:
        element.pop("id")
    return boards, menu_paths, components, data_by_data_set


def create_bar_chart(shimoku_client):
    data = [
        {"date": dt.date(2021, 1, 1), "x": 50000000, "y": 5},
        {"date": dt.date(2021, 1, 2), "x": 60000000, "y": 5},
        {"date": dt.date(2021, 1, 3), "x": 40000000, "y": 5},
        {"date": dt.date(2021, 1, 4), "x": 70000000, "y": 5},
        {"date": dt.date(2021, 1, 5), "x": 30000000, "y": 5},
    ]
    shimoku_client.plt.bar(
        data=data,
        x="date",
        y=["x", "y"],
        x_axis_name="Date",
        y_axis_name="Revenue",
        order=0,
        rows_size=2,
        cols_size=12,
    )


class TestCodeGen(unittest.TestCase):
    def setUp(self):
        self.shimoku_client = initiate_shimoku()
        self.workspace_id = getenv("BUSINESS_ID")
        self.shimoku_client.set_workspace(self.workspace_id)

    def test_code_gen(self):
        if not self.shimoku_client.playground or mock:
            # It would take too long to complete the test
            return
        self.shimoku_client.set_workspace()
        clear_workspace(self.shimoku_client)
        main()
        try:
            self.shimoku_client.pop_out_of_board()
        except BoardError:
            pass
        self.shimoku_client.disable_caching()
        boards, menu_paths, components, data = get_workspace_contents(
            self.shimoku_client
        )
        temp_dir = tempfile.mkdtemp()
        subprocess.run(
            [
                "shimoku",
                "persist",
                "generate-code",
                "--output-path",
                temp_dir,
                "--universe-id",
                "local",
                "--workspace",
                "local",
            ]
        )
        clear_workspace(self.shimoku_client)
        subprocess.run(["python", f"{temp_dir}/execute_workspace_local.py"], check=True)
        (
            first_generation_boards,
            first_generation_menu_paths,
            first_generation_components,
            first_generation_data,
        ) = get_workspace_contents(self.shimoku_client)
        shutil.rmtree(temp_dir)
        temp_dir = tempfile.mkdtemp()
        subprocess.run(
            [
                "shimoku",
                "persist",
                "generate-code",
                "--output-path",
                temp_dir,
                "--universe-id",
                "local",
                "--workspace",
                "local",
            ]
        )
        subprocess.run(["python", f"{temp_dir}/execute_workspace_local.py"], check=True)
        (
            second_generation_boards,
            second_generation_menu_paths,
            second_generation_components,
            second_generation_data,
        ) = get_workspace_contents(self.shimoku_client)
        clear_workspace(self.shimoku_client)
        self.shimoku_client.run()
        shutil.rmtree(temp_dir)
        results = {
            "orig-first": (
                get_diff_percentage(boards, first_generation_boards),
                get_diff_percentage(menu_paths, first_generation_menu_paths),
                get_diff_percentage(components, first_generation_components),
                get_diff_percentage(data, first_generation_data),
            ),
            "orig-second": (
                get_diff_percentage(boards, second_generation_boards),
                get_diff_percentage(menu_paths, second_generation_menu_paths),
                get_diff_percentage(components, second_generation_components),
                get_diff_percentage(data, second_generation_data),
            ),
            "first-second": (
                get_diff_percentage(first_generation_boards, second_generation_boards),
                get_diff_percentage(
                    first_generation_menu_paths, second_generation_menu_paths
                ),
                get_diff_percentage(
                    first_generation_components, second_generation_components
                ),
                get_diff_percentage(first_generation_data, second_generation_data),
            ),
        }
        print(json.dumps(results, indent=4))
        assert all(
            all([value == 0 for value in results_list])
            for results_list in results.values()
        )

    def test_commit(self):
        if self.shimoku_client.playground or mock:
            # Test commit only in cloud
            return
        local_shimoku_client = Client(verbosity="INFO")
        local_shimoku_client.disable_caching()
        local_shimoku_client.set_workspace()
        clear_workspace(local_shimoku_client)
        local_shimoku_client.set_board("Commit-board")
        local_shimoku_client.set_menu_path("Commit-menu-path")
        create_bar_chart(local_shimoku_client)
        local_shimoku_client.pop_out_of_board()
        (
            local_boards,
            local_menu_paths,
            local_components,
            local_data,
        ) = get_workspace_contents(local_shimoku_client)
        clear_workspace(self.shimoku_client)
        subprocess.run(
            [
                "shimoku",
                "persist",
                "commit",
                "--universe-id",
                "local",
                "--workspace",
                "local",
                "--target-access-token",
                self.shimoku_client.access_token,
                "--target-universe-id",
                self.shimoku_client.universe_id,
                "--target-workspace",
                self.workspace_id,
                "--target-environment",
                self.shimoku_client.environment,
            ]
        )
        self.shimoku_client.disable_caching()
        self.shimoku_client.set_workspace(self.workspace_id)
        boards, menu_paths, components, data = get_workspace_contents(
            self.shimoku_client
        )
        results = {
            "boards": (get_diff_percentage(local_boards, boards)),
            "menu_paths": (get_diff_percentage(local_menu_paths, menu_paths)),
            "components": (get_diff_percentage(local_components, components)),
            "data": (get_diff_percentage(local_data, data)),
        }
        print(json.dumps(results, indent=4))
        print(local_menu_paths)
        print(menu_paths)
        assert all([value == 0 for value in results.values()])

    def test_pull(self):
        if self.shimoku_client.playground or mock:
            # Test pull only in cloud
            return
        self.shimoku_client.set_workspace(self.workspace_id)
        clear_workspace(self.shimoku_client)
        self.shimoku_client.set_board("Pull-board")
        self.shimoku_client.set_menu_path("Pull-menu-path")
        create_bar_chart(self.shimoku_client)
        self.shimoku_client.pop_out_of_board()
        self.shimoku_client.disable_caching()
        tpa_boards, tpa_menu_paths, tpa_components, tpa_data = get_workspace_contents(
            self.shimoku_client
        )

        local_shimoku_client = Client(verbosity="INFO")
        local_shimoku_client.set_workspace()
        clear_workspace(local_shimoku_client)
        subprocess.run(
            [
                "shimoku",
                "persist",
                "pull",
                "--universe-id",
                "local",
                "--workspace",
                "local",
                "--origin-access-token",
                self.shimoku_client.access_token,
                "--origin-universe-id",
                self.shimoku_client.universe_id,
                "--origin-workspace",
                self.workspace_id,
                "--origin-environment",
                self.shimoku_client.environment,
            ]
        )
        local_shimoku_client.disable_caching()
        (
            local_boards,
            local_menu_paths,
            local_components,
            local_data,
        ) = get_workspace_contents(local_shimoku_client)

        clear_workspace(self.shimoku_client)
        clear_workspace(local_shimoku_client)

        results = {
            "boards": (get_diff_percentage(tpa_boards, local_boards)),
            "menu_paths": (get_diff_percentage(tpa_menu_paths, local_menu_paths)),
            "components": (get_diff_percentage(tpa_components, local_components)),
            "data": (get_diff_percentage(tpa_data, local_data)),
        }
        print(json.dumps(results, indent=4))
        print(tpa_menu_paths)
        print(local_menu_paths)
        assert all([value == 0 for value in results.values()])



from shimoku import Client
from .data.suburst import sunburst_data
from .data.tree import tree_data


def treemap(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "treemap-test")
    # Using the data set from the tree test
    shimoku_client.plt.treemap(data=tree_data, order=0)
    shimoku_client.plt.treemap(data=sunburst_data, order=1)



from shimoku import Client
from .data.line_with_confidence_area import confidence_data


def line_with_confidence_area(shimoku_client: Client):
    shimoku_client.set_menu_path("test-free-echarts", "Line with confidence area")

    shimoku_client.plt.line_with_confidence_area(
        data=confidence_data,
        order=0,
        title="Confidence Band Chart",
        x="date",
        y="value",
        lower="l",
        upper="u",
        x_axis_name="Date",
        y_axis_name="Value",
        percentages=True,
        variant="clean",
    )



from shimoku import Client
from .data.suburst import sunburst_data


def sunburst(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "sunburst")
    # Using the data set from the tree test
    shimoku_client.plt.sunburst(data="tree_data", order=0)
    shimoku_client.plt.sunburst(data=sunburst_data, order=1)



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame
from dataclasses import dataclass
import shimoku


class AIFunctionMethodsHeader:
    """None"""

    def create_model(
        self,
        model_name: str,
        model: bytes,
        metadata: Optional[dict] = None,
    ):
        """Create a model
        :param model_name: Name of the model
        :param model: Model to be uploaded
        :param metadata: Metadata of the model
        """
        pass

    def create_output_files(
        self,
        files: dict,
        model_name: Optional[str] = None,
    ):
        """Create output files of a ai_function
        :param files: Files to be uploaded
        :param model_name: Name of the model used
        """
        pass

    def get_model(
        self,
        model_name: str,
    ) -> tuple:
        """Get a model
        :param model_name: Name of the model
        :return: The model if it exists
        """
        pass


class AILayerHeader:
    """None"""

    def check_for_private_access(
        self,
        ai_function_id: str,
        run_id: str,
    ):
        """Check if the credentials allow for private ai_function methods"""
        pass

    def create_input_files(
        self,
        input_files: dict,
        force_overwrite: bool = False,
    ):
        """Create input files for a ai_function
        :param input_files: dictionary of input files to create
        :param force_overwrite: Force the overwrite of the files
        """
        pass

    def delete_input_file(
        self,
        file_name: str,
    ):
        """Delete an input file
        :param file_name: Name of the file to delete
        """
        pass

    def delete_model(
        self,
        model_name: str,
    ):
        """Delete a model
        :param model_name: Name of the model
        """
        pass

    def generic_execute(
        self,
        ai_function: str,
        version: Optional[str] = None,
        universe_api_key: Optional[str] = None,
        **kwargs,
    ) -> str:
        """Execute a generic ai_function
        :param ai_function: Name of the ai_function to execute
        :param version: Version of the ai_function to execute
        :param universe_api_key: API key of the universe
        :param params: Parameters to be passed to the ai_function
        """
        pass

    def get_available_input_files(
        self,
    ) -> list:
        """Get available input files for a ai_function"""
        pass

    def get_last_executions_with_logs(
        self,
        ai_function: str,
        version: Optional[str] = None,
        how_many: int = 1,
    ):
        """Get the logs of the executions of a ai_function
        :param ai_function: Name of the ai_function to execute
        :param version: Version of the ai_function to execute
        :param how_many: Number of executions to get
        :return: The logs of the ai_function
        """
        pass

    def get_output_file_objects(
        self,
        run_id: str,
        file_name: Optional[str] = None,
    ) -> Union[tuple[bytes, dict], dict[str, tuple[bytes, dict]]]:
        """Get an output file object for a given ai_function and run id
        :param run_id: ID of the executed run
        :param file_name: Name of the file to get
        """
        pass

    def get_output_files_by_ai_function(
        self,
        ai_function: str,
        version: Optional[str] = None,
        file_name: Optional[str] = None,
        get_objects: bool = False,
    ):
        """Get output files for a generic ai_function
        :param ai_function: Name of the ai_function to execute
        :param version: Version of the ai_function to execute
        :return: The output files if they exist
        :param file_name: Name of the file to get
        :param get_objects: Get the file objects instead of the file metadata
        """
        pass

    def get_output_files_by_model(
        self,
        model_name: str,
        file_name: Optional[str] = None,
        get_objects: bool = False,
    ):
        """Get output files for the executions of ai_functions with a given model
        :param model_name: Name of the model to use
        :param file_name: Name of the file to get
        :param get_objects: Get the file objects instead of the file metadata
        :return: dictionary of output files by the execution identifier
        """
        pass

    def get_private_ai_function_methods(
        self,
    ) -> AIFunctionMethodsHeader:
        """Get the private ai_function methods if the credentials allow for it"""
        pass



from typing import Optional, TYPE_CHECKING, Type

import asyncio
from copy import deepcopy

from shimoku.api.base_resource import Resource
from shimoku.api.resources.data_set import DataSet, Mapping

if TYPE_CHECKING:
    from shimoku.api.resources.app import App

import logging

logger = logging.getLogger(__name__)


class Report(Resource):
    """Report resource class"""

    _module_logger = logger
    resource_type = "report"
    plural = "reports"
    alias_field = "properties/hash"

    report_type = None
    possible_values = {}
    default_properties = {
        "hash": None,
    }

    class ReportDataSet(Resource):
        _module_logger = logger
        resource_type = "reportDataSet"
        plural = "reportDataSets"

        def __init__(
            self,
            parent: "Report",
            uuid: Optional[str] = None,
            db_resource: Optional[dict] = None,
        ):
            params = dict(
                reportId=parent["id"],
                dataSetId="",
                properties={},
            )

            super().__init__(
                parent=parent,
                uuid=uuid,
                db_resource=db_resource,
                params=params,
                check_params_before_creation=["reportId", "dataSetId"],
                params_to_serialize=["properties"],
            )

        async def delete(self):
            """Delete the report dataset"""
            return await self._base_resource.delete()

        async def update(self) -> bool:
            """Update the report dataset"""
            return await self._base_resource.update()

    def __init__(
        self,
        parent: "App",
        uuid: Optional[str] = None,
        db_resource: Optional[dict] = None,
    ):
        params = dict(
            title=None,
            path=None,
            pathOrder=None,
            reportType=self.report_type,
            order=0,
            sizeColumns=12,
            sizeRows=3,
            sizePadding="0,0,0,0",
            bentobox={},
            properties=deepcopy(self.default_properties),
            dataFields={},
            chartData=[],
            # subscribed=False,
            smartFilters=[],
        )

        super().__init__(
            parent=parent,
            uuid=uuid,
            db_resource=db_resource,
            params=params,
            children=[Report.ReportDataSet],
            check_params_before_creation=["order"],
            params_to_serialize=[
                "properties",
                "dataFields",
                "chartData",
                "bentobox",
                "smartFilters",
            ],
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other: "Report"):
        return super().__eq__(other)

    def __new__(
        cls,
        parent: "App",
        uuid: Optional[str] = None,
        db_resource: Optional[dict] = None,
    ):
        if uuid:
            return super().__new__(cls)
        if cls is Report:
            if db_resource is None:
                raise ValueError(
                    "You must provide a db_resource to create a Report instance"
                )
            if db_resource["reportType"] == "TABS":
                from .reports.tabs_group import TabsGroup

                return TabsGroup(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource["reportType"] == "MODAL":
                from .reports.modal import Modal

                return Modal(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource["reportType"] == "INDICATOR":
                from .reports.charts.indicator import Indicator

                return Indicator(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource["reportType"] == "ECHARTS2":
                from .reports.charts.echart import EChart

                return EChart(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource["reportType"] == "IFRAME":
                from .reports.charts.iframe import IFrame

                return IFrame(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource["reportType"] == "HTML":
                from .reports.charts.html import HTML

                return HTML(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource["reportType"] == "TABLE":
                from .reports.charts.table import Table

                return Table(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource["reportType"] == "ANNOTATED_ECHART":
                from .reports.charts.annotated_chart import AnnotatedEChart

                return AnnotatedEChart(
                    parent=parent, uuid=uuid, db_resource=db_resource
                )
            elif db_resource["reportType"] == "BUTTON":
                from .reports.charts.button import Button

                return Button(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource["reportType"] == "FORM":
                from .reports.charts.input_form import InputForm

                return InputForm(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource["reportType"] == "FILTERDATASET":
                from .reports.filter_data_set import FilterDataSet

                return FilterDataSet(parent=parent, uuid=uuid, db_resource=db_resource)
            elif db_resource["reportType"] in [
                "INDICATORS",
                "MULTIFILTER",
                "ECHARTS",
                None,
            ]:
                from .reports.unsupported import Unsupported

                return Unsupported(parent=parent, uuid=uuid, db_resource=db_resource)
            else:
                raise ValueError(f'Unknown report type {db_resource["reportType"]}')
        else:
            return super().__new__(cls)

    async def delete(self):
        """Delete the report"""
        return await self._base_resource.delete()

    async def update(self) -> bool:
        """Update the report"""
        return await self._base_resource.update()

    def set_properties(self, **properties):
        """Set the properties of the report without saving it to the server ç
        :param properties: the properties to set"""
        for property_name, value in properties.items():
            if property_name not in self.default_properties.keys():
                raise ValueError(
                    f"Property {property_name} is not a possible property for {self.report_type}"
                )
            if (
                property_name in self.possible_values
                and value not in self.possible_values[property_name]
            ):
                raise ValueError(
                    f"Value {value} is not a possible value for property {property_name}"
                )

        self["properties"].update(properties)
        self._base_resource.changed_params.add("properties")

    async def change_report_type(self, report_class: Type["Report"]):
        """change the report type of the report"""
        logger.info(
            f"Changing component type from {self.report_type} to {report_class.report_type}"
        )
        if self.report_type == report_class.report_type:
            return
        await self.delete_report_data_sets()
        self["reportType"] = report_class.report_type
        r_hash = self["properties"].get("hash")
        self["properties"] = deepcopy(report_class.default_properties)
        self["properties"]["hash"] = r_hash
        self["dataFields"] = {}
        self["chartData"] = []
        self.__class__ = report_class

    # ReportDataSet methods
    async def create_report_data_set(
        self,
        mapping_data_set_sort: tuple[Optional[Mapping], DataSet, Optional[dict]],
        properties: Optional[dict] = None,
    ) -> "ReportDataSet":
        """Update the report dataset
        :param mapping_data_set_sort: a tuple of mapping, data_set, sort
        :param properties: the properties to set
        :return: the updated report dataset
        """
        mapping, data_set, sort = mapping_data_set_sort
        data_set_id = data_set["id"]
        properties = properties if properties is not None else {}
        properties["mapping"] = mapping
        properties["sort"] = sort
        params = dict(
            dataSetId=data_set_id,
            properties=properties,
        )
        report_data_set = await self._base_resource.create_child(
            self.ReportDataSet, **params
        )
        return report_data_set

    async def get_report_dataset(self, uuid: Optional[str] = None) -> "ReportDataSet":
        """Get a report dataset
        :param uuid: the uuid of the report dataset to get
        :return: the report dataset
        """
        return await self._base_resource.get_child(self.ReportDataSet, uuid=uuid)

    async def get_report_data_sets(self) -> list["ReportDataSet"]:
        """Get all the report datasets of the report"""
        return await self._base_resource.get_children(self.ReportDataSet)

    async def delete_report_dataset(self, uuid: Optional[str] = None) -> bool:
        """Delete a report dataset
        :param uuid: the uuid of the report dataset to delete
        :return: True if the report dataset was deleted, False otherwise
        """
        return await self._base_resource.delete_child(self.ReportDataSet, uuid=uuid)

    async def delete_report_data_sets(self, log: bool = False):
        """Delete the datasets and links that are not in use anymore.
        :param log: if True, log the actions_execution"""
        report_data_sets = await self.get_report_data_sets()
        await asyncio.gather(
            *[
                self._base_resource.delete_child(self.ReportDataSet, uuid=rds["id"])
                for rds in report_data_sets
            ]
        )
        if log:
            logger.info(
                f"Deleted {len(report_data_sets)} component data set links from component at {str(self)}"
            )



pie_data = [
    {"value": 1048, "name": "Search Engine"},
    {"value": 735, "name": "Direct"},
    {"value": 580, "name": "Email"},
    {"value": 484, "name": "Union Ads"},
    {"value": 300, "name": "Video Ads"},
]



from shimoku import Client
from .data.input_form import input_form_data


def get_input_forms(shimoku_client: Client):
    shimoku_client.set_menu_path("Input forms", "get input forms")
    shimoku_client.plt.input_form(order=0, options=input_form_data)
    rs: list[dict] = shimoku_client.plt.get_input_forms()
    print(rs)
    assert rs
